/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.anweisung;

import java.util.TimerTask;

import de.bsvrz.ibv.uda.interpreter.AnweisungsBlock;
import de.bsvrz.ibv.uda.interpreter.daten.SkriptKontext;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.Fehler;
import de.bsvrz.ibv.uda.interpreter.daten.zeit.ZeitDauer;

/**
 * die Repräsenation der nebenläufigen Aktivität: <b>Zyklus</b> innerhalb eines
 * UdaInterpreters.<br>
 * Die Klasse implementiert die Javaklasse TimerTask und wird als Aufgabe im
 * Timer des Interpreter eingetragen. Nach Ablauf der Zykluszeit werden die
 * zugeordneten Anweisungen ausgeführt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class Zyklus extends UdaAktivitaet {

	/**
	 * die timergesteuerte Aufgabe.
	 *
	 * @author peuker
	 */
	public class ZyklusTask extends TimerTask {

		/**
		 * der Zyklus, dessen Anweisungen zyklisch ausgeführt werden sollen.
		 */
		private final Zyklus zyklus;

		/**
		 * Konstruktor.
		 *
		 * @param zyklus
		 *            der Zyklus, zu dem die Aufgabe gehört
		 */
		public ZyklusTask(final Zyklus zyklus) {
			this.zyklus = zyklus;
		}

		/**
		 * führt die Anweisungen des Zyklus aus. {@inheritDoc}.
		 *
		 * @see java.util.TimerTask#run()
		 */
		@Override
		public void run() {
			zyklus.ausfuehren();
		}
	}

	/**
	 * die zyklisch auszuführenden Anweisungen.
	 */
	private final AnweisungsBlock anweisungen;

	/**
	 * die zyklisch auszuführende Aufgabe.
	 */
	private final TimerTask zyklusTask;

	/**
	 * Konstruktor.
	 *
	 * @param kontext
	 *            der Kontext des aufrufenden Anweisungsblocks
	 * @param name
	 *            der Name, unter dem die nebenläufige Anweisung im umgebenden
	 *            Kontext eingetragen wird
	 * @param intervall
	 *            das Intervall für die Ausführung der Anweisungen
	 * @param anweisungen
	 *            die auszuführenden Anweisungen
	 */
	public Zyklus(final SkriptKontext kontext, final String name, final ZeitDauer intervall,
			final AnweisungsBlock anweisungen) {
		super(kontext, name);
		this.anweisungen = anweisungen;
		zyklusTask = new ZyklusTask(this);
		kontext.getInterpreter().getTimer().scheduleAtFixedRate(zyklusTask, 0, intervall.getMilliSekunden());
		setRunning(true);
		kontext.getInterpreter().addAktivitaet(this);
	}

	/**
	 * führt den Anweisungsblock der Anweisung aus.<br>
	 * Es werden alle Anweisungen abgearbeitet. Tritt eine unbehandelte Ausnahme
	 * auf, wird das Fehlerobjekt der Aktivität gesetzt und die Aktivität
	 * beendet.
	 */
	public void ausfuehren() {
		try {
			if (anweisungen.interpret(getKontext()) instanceof RueckgabeAnweisung) {
				beenden();
			}
		} catch (final RuntimeException e) {
			setFehler(new Fehler(e.getMessage()));
			beenden();
		}
	}

	/**
	 * beendet die Aktivität.<br>
	 * Der Status der Aktivität wird auf inaktiv gesetzt und der ausführende
	 * Interpreter benachrichtigt. Die zyklisch auszuführende TimerTask wird
	 * beendet.<br>
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.anweisung.UdaAktivitaet#beenden()
	 */
	@Override
	public void beenden() {
		zyklusTask.cancel();
		setRunning(false);
		synchronized (getKontext().getInterpreter()) {
			getKontext().getInterpreter().notifyAll();
		}
	}
}
