/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.uda.interpreter.daten.dav.BestaetigterDavSender;
import de.bsvrz.ibv.uda.interpreter.daten.dav.DavDaten;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.KonfigurationsFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.ibv.uda.interpreter.daten.zeit.ZeitDauer;
import de.bsvrz.ibv.uda.interpreter.daten.zeit.Zeitstempel;

/**
 * die Repräsentation einer Anmeldung für den Empfang oder den Versand von Daten
 * beim Datenverteiler.<br>
 * Die Anmeldung kann als Quelle, Senke, Sender oder Empfänger erfolgen. Die
 * Anmeldung ist beschrieben durch ein Systemobjekt für das die Anmeldung
 * erfolgte sowie eine zugehörige Attributgruppe mit einem passenden Aspekt.<br>
 * Zusätzlich wird eine Timeoutzeit gespeichert, die bei Empfängeranmeldungen
 * zur Ausführung des Fehlerbehandlungsblockes in einer Anmeldeanweisung führen
 * kann.<br>
 * Für Sende- und Quellenanmeldungen steht außerdem der Sendeanforderungsstatus
 * zur Verfügung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class AtgAnmeldungsEintrag implements AnmeldungsEintrag {
	/**
	 * das Systemobjekt.
	 */
	private SystemObject objekt;

	/**
	 * die Attributgruppe.
	 */
	private AttributeGroup attributGruppe;

	/**
	 * der Aspekt.
	 */
	private Aspect aspekt;

	/**
	 * das Timeoutintervall (0, wenn keine Überwachung erfolgen soll).
	 */
	private long timeOut;

	/**
	 * der letzte Aktualisierungszeitpunkt der Daten für diese Anmeldung.
	 */
	private long aktualisierung = System.currentTimeMillis();

	/**
	 * der Zustand der Sendeanforderung.
	 */
	private byte requestStatus = ClientSenderInterface.STOP_SENDING;

	/**
	 * die verwendete Datenverteilerverbindung.
	 */
	private final ClientDavInterface verbindung;

	/**
	 * Konstruktor.<br>
	 * Die Daten für das Systemobjekt, die Attributgruppe und den Aspekt können
	 * entweder direkt als das entsprechende Datenverteilerobjekt,als
	 * udainternes gekapseltes Datenverteilerobjekt oder als PID (Zeichenkette)
	 * übergeben werden. Wird ein ungültiges Objekt üebrgeben bzw. kann aus dem
	 * übergebenen Objekt nicht das erforderliche Objekt für die Anmeldung
	 * ermittelt werden wird ein entsprechender Fehler ausgelöst.<br>
	 * Die eigentliche Anmeldung erfolgt mit der Funktion
	 * {@link #anmelden(AnmeldeTyp, Anmeldung)}.
	 *
	 * @param dav
	 *            die Datenverteilerverbindung
	 * @param objObjekt
	 *            das Objekt, das das Systemobjekt definiert
	 * @param atgObjekt
	 *            das Objekt, das die Attributgruppe definiert
	 * @param aspObjekt
	 *            das Objekt, das den Aspekt definiert
	 * @param dauer
	 *            das Timeoutintervall
	 */
	public AtgAnmeldungsEintrag(final ClientDavInterface dav, final Object objObjekt, final Object atgObjekt,
			final Object aspObjekt, final ZeitDauer dauer) {

		this.verbindung = dav;

		if (dauer != null) {
			timeOut = dauer.getMilliSekunden();
		}

		try {
			this.objekt = (SystemObject) Anmeldung.getDavObject(dav, objObjekt, SystemObject.class);
			this.attributGruppe = (AttributeGroup) Anmeldung.getDavObject(dav, atgObjekt, AttributeGroup.class);
			this.aspekt = (Aspect) Anmeldung.getDavObject(dav, aspObjekt, Aspect.class);
		} catch (@SuppressWarnings("unused") final ArgumentFehler e) {
			// erzeugen der entsprechenden Fehler erfolgt im nächsten Schritt
		}

		if (objekt == null) {
			throw new KonfigurationsFehler(UdaFehlerSubtyp.OBJEKT);
		}
		if (attributGruppe == null) {
			throw new KonfigurationsFehler(UdaFehlerSubtyp.ATTRIBUTGRUPPE);
		}
		if (aspekt == null) {
			throw new KonfigurationsFehler(UdaFehlerSubtyp.ASPEKT);
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Daten werden entsprechend der definierten Datenspezifikation vom
	 * Datenverteiler abgemeldet.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.AnmeldungsEintrag#abmelden(de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.AnmeldeTyp,
	 *      de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.Anmeldung)
	 */
	@Override
	public void abmelden(final AnmeldeTyp typ, final Anmeldung anmeldung) {

		switch (typ) {
		case SENKE:
		case EMPFAENGER:
			anmeldung.getConnection().unsubscribeReceiver(anmeldung, objekt,
					new DataDescription(attributGruppe, aspekt));
			break;
		case SENDER:
		case QUELLE:
			BestaetigterDavSender.getSender(anmeldung.getConnection()).abmelden(objekt,
					new DataDescription(attributGruppe, aspekt));
			break;
		default:
			throw new UnsupportedOperationException("Der Anmeldetyp: " + typ
					+ " wird nicht unterstützt.");
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Anmeldung erfolgt gemäß der Datenspezifikation des Anmeldeinetrags.
	 * Empfänger und Quellen werden mit der ReceiverOption "delayed" angemeldet,
	 * d.h. auch für nachgelieferte und unveränderte Daten.<br>
	 * Eine mehrfache Anmeldung mit identischer Datenspezifikation als Sender
	 * führt zu einem entsprechenden Fehler.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.AnmeldungsEintrag#anmelden(de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.AnmeldeTyp,
	 *      de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.Anmeldung)
	 */
	@Override
	public void anmelden(final AnmeldeTyp typ, final Anmeldung anmeldung) {

		switch (typ) {
		case EMPFAENGER:
			anmeldung.getConnection().subscribeReceiver(anmeldung, objekt, new DataDescription(attributGruppe, aspekt),
					ReceiveOptions.delayed(), ReceiverRole.receiver());
			break;
		case SENKE:
			anmeldung.getConnection().subscribeReceiver(anmeldung, objekt, new DataDescription(attributGruppe, aspekt),
					ReceiveOptions.delayed(), ReceiverRole.drain());
			break;
		case SENDER:
			BestaetigterDavSender.getSender(anmeldung.getConnection()).anmelden(objekt,
					new DataDescription(attributGruppe, aspekt), false);
			break;
		case QUELLE:
			BestaetigterDavSender.getSender(anmeldung.getConnection()).anmelden(objekt,
					new DataDescription(attributGruppe, aspekt), true);
			break;
		default:
			throw new UnsupportedOperationException("Der Anmeldetyp: " + typ
					+ " wird nicht unterstützt.");
		}
	}

	/**
	 * liefert den Status der Sendeanforderungssteuerung.
	 *
	 * @return den Status
	 */
	protected byte getRequestStatus() {
		return requestStatus;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Folgende Strukturelemente werden unterstützt:
	 * <ul>
	 * <li>objekt: liefert das Systemobjekt der Anmeldung</li>
	 * <li>attributgruppe: liefert die Attributgruppe der Anmeldung</li>
	 * <li>aspekt: liefert den Aspekt der Anmeldung</li>
	 * <li>timeout: liefert den Tineoutwert der Anmeldung</li>
	 * <li>aktualisierung: liefert den Zeitstempel der letzten Aktualisierung
	 * </li>
	 * </ul>
	 *
	 * @see #setStrukturElement(String, Object)
	 */
	@Override
	public Object getStrukturElement(final String name) {
		Object ergebnis = null;
		if ("objekt".equals(name)) {
			ergebnis = DavDaten.erzeugeObjekt(verbindung, objekt);
		} else if ("attributgruppe".equals(name)) {
			ergebnis = DavDaten.erzeugeObjekt(verbindung, attributGruppe);
		} else if ("aspekt".equals(name)) {
			ergebnis = DavDaten.erzeugeObjekt(verbindung, aspekt);
		} else if ("timeout".equals(name)) {
			ergebnis = timeOut;
		} else if ("aktualisierung".equals(name)) {
			ergebnis = new Zeitstempel(aktualisierung);
		}
		return ergebnis;
	}

	/**
	 * liefert den Timeoutwert der Anmeldung in Millisekunden.<br>
	 * Der Wert 0 steht für "Keine Überwachung".
	 *
	 * @return den Wert
	 */
	long getTimeOut() {
		return timeOut;
	}

	/**
	 * ermittelt, ob die Timeoutzeit für die Anmeldung abgelaufen ist.
	 *
	 * @param testZeit
	 *            der Zeitpunkt (in Millisekunden) für den der Test ausgeführt
	 *            werden soll.
	 * @return wahr, wenn der Timeout abgelaufen ist
	 */
	public boolean isTimedOut(final long testZeit) {
		if (timeOut > 0) {
			if ((testZeit - aktualisierung) > timeOut) {
				return true;
			}
		}
		return false;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung.AnmeldungsEintrag#matches(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.config.AttributeGroup,
	 *      de.bsvrz.dav.daf.main.config.Aspect)
	 */
	@Override
	public boolean matches(final SystemObject sysObj, final AttributeGroup atg, final Aspect asp) {
		boolean result = false;
		if (objekt.equals(sysObj) && this.attributGruppe.equals(atg) && this.aspekt.equals(asp)) {
			result = true;
		}
		return result;
	}

	/**
	 * setzt den Zeitpunkt der letzten Aktualisierung des Eintrags,<br>
	 * Der Zeitpunkt wird nur übernommen, wenn er größer als der zuletzt
	 * gespeicherte ist.
	 *
	 * @param wert
	 *            der Aktulisierungszeitpunkt in Millisekunden
	 */
	public void setAktualisierung(final long wert) {
		if (wert > aktualisierung) {
			aktualisierung = wert;
		}
	}

	/**
	 * setzt den Status der Sendeanforderungssteuerung.
	 *
	 * @param status
	 *            der Status
	 */
	protected void setRequestState(final byte status) {
		requestStatus = status;

	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Attribute der Anmeldung können nicht von außen geändert werden. Jeder
	 * Aufruf dieser Funktion führt daher zu einem ArgumentFehler.
	 *
	 * @see #getStrukturElement(String)
	 */
	@Override
	public void setStrukturElement(final String name, final Object daten) {
		throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE,
				"Anmeldungen können nicht per Elementzugriff verändert werden");

	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Ausgabe der Anmeldung als Zeichenkette erfolgt in der Form:
	 *
	 * <pre>
	 * (&lt;objekt&gt;,&lt;attributgruppe&gt;,&lt;aspekt&gt;)
	 * </pre>
	 *
	 * ,<br>
	 * wobei für Objekt, Attributgruppe un den Aspekt jeweils die PID oder der
	 * Name oder die ID ausgegeben werden.
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer("(");
		buffer.append(objekt.getPidOrNameOrId());
		buffer.append(',');
		buffer.append(attributGruppe.getPidOrNameOrId());
		buffer.append(',');
		buffer.append(aspekt.getPidOrNameOrId());
		buffer.append(')');
		return buffer.toString();
	}
}
