/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.anweisung.iteration;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.ibv.uda.interpreter.AnweisungsBlock;
import de.bsvrz.ibv.uda.interpreter.UdaInterpreter;
import de.bsvrz.ibv.uda.interpreter.anweisung.AbstractAnweisung;
import de.bsvrz.ibv.uda.interpreter.anweisung.RueckgabeAnweisung;
import de.bsvrz.ibv.uda.interpreter.daten.SkriptKontext;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehler;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.bitctrl.interpreter.NichtWert;

/**
 * die Repräsentation einer Iterationsanweisung innerhalb eines Uda-Skripts.<br>
 * Eine Iterationsanweisung enthält eine Iterationsbedingung und einen
 * Anweisungsblock. Der Anweisungsblock wird solange ausgeführt, bis die
 * Bedingung nicht mehr gültig ist.<br>
 * Uda unterstützt folgende Iterationen:
 * <ul>
 * <li>Iteration per Laufvariable über eine Menge</li>
 * <li>Iteration mit einer Bedingung (entspricht einer WHILE-Schleife)</li>
 * <li>Iteration mit numerischen Laufvariable (entspricht einer FOR-Schleife mit
 * ganzzahligem Laufindex)</li>
 * </ul>
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class IterationsAnweisung extends AbstractAnweisung {

	/**
	 * die vor jeder Ausführung des Anweisungsblocks auszuwertende Bedingung.
	 */
	private final IterationsBedingung bedingung;

	/**
	 * der Anweisungsblock.
	 */
	private final AnweisungsBlock anweisungen;

	/**
	 * Konstruktor.
	 *
	 * @param zeile
	 *            die Nummer der Zeile, in der die Anweisung definiert ist
	 * @param skript
	 *            der name des Skripts
	 * @param bedingung
	 *            die Bedingung unter der die Anweisungen ausgeführt werden
	 * @param anweisungen
	 *            die auszuführenden Anweisungen
	 */
	public IterationsAnweisung(final int zeile, final String skript, final IterationsBedingung bedingung,
			final AnweisungsBlock anweisungen) {
		super(zeile, skript);
		this.bedingung = bedingung;
		this.anweisungen = anweisungen;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck#getNachfolger()
	 */
	@Override
	public List<Ausdruck> getNachfolger() {
		final List<Ausdruck> liste = new ArrayList<>();
		liste.add(bedingung);
		liste.add(anweisungen);
		return liste;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion erzeugt einen lokalen Kontext, in den alle Variablen des
	 * übergebenden Kontextes übernommen werden, um ein überschreiben einer
	 * Variable des umgebenden Kontextes durch eine Laufvariable zu vermeiden.
	 * Die Bedingung wird geprüft, wenn diese den Wahrheitswert "wahr" liefert,
	 * wird der Anweisungsblock ausgeführt. Dies wird solange wiederholt bis die
	 * Bedingung nicht mehr gültig ist.<br>
	 * Tritt bei der Auswertung der Bedingung oder der Ausführung des
	 * Anweisungsblockes ein unbehandelter Fehler auf, wird die Ausführung der
	 * Iteration ebenfalls unterbrochen und die Ausnahme weitergereicht. Am Ende
	 * der Ausführung der Anweisung werden alle Daten, bis auf die Daten der
	 * Laufvariable in den umgebenden Kontext zurückübertragen.<br>
	 * Die Anweisung liefert immer den Wert undefiniert zurück.
	 *
	 * @see de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck#interpret(de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext)
	 */
	@Override
	public final Object interpret(final Kontext kontext) {
		Object result = null;
		final SkriptKontext iterationsKontext = new SkriptKontext((SkriptKontext) kontext, true);
		UdaFehler fehler = null;
		final UdaInterpreter interpreter = iterationsKontext.getInterpreter();
		try {
			bedingung.reset();
			while ((interpreter.isRunning() && bedingung.isGueltig(iterationsKontext))) {
				result = anweisungen.interpret(iterationsKontext);
				if (result instanceof RueckgabeAnweisung) {
					break;
				}
			}
		} catch (final InterpreterException e) {
			fehler = erzeugeAnweisungsFehler(e);
		}

		final String variable = bedingung.getLaufVariable();
		for (final String key : kontext.getVariablen()) {
			if (!key.equals(variable)) {
				final Object obj = iterationsKontext.get(key);
				if (obj != null) {
					kontext.set(key, obj);
				}
			}
		}

		if (fehler != null) {
			throw fehler;
		}

		if (result == null) {
			result = NichtWert.NICHTWERT;
		}
		return result;
	}
}
