/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.daten.java;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import de.bsvrz.ibv.uda.interpreter.Kopierbar;
import de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.Fehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation eines externen eingebunden Java-Objekts.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class JavaObjekt implements Struktur, Kopierbar {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Qualität der Übereinstimmung bei der Ermiitlung einer passenden
	 * Methode eines Java-Objekts.
	 *
	 * @author peuker
	 */
	private enum MatchingQuality {
		/**
		 * Methode passt genau.
		 */
		FULL,

		/**
		 * die Argumente können auf die Argumenttypen der Methide angepasst
		 * werden.
		 */
		NEAR,

		/**
		 * die Methode passt nicht.
		 */
		NONE;
	}

	/**
	 * das eingebundene Java-Objekt.
	 */
	private final Object objekt;

	/**
	 * Konstruktor.<br>
	 *
	 * @param objekt
	 *            das zu Grunde liegende Java-Object.
	 */
	public JavaObjekt(final Object objekt) {
		this.objekt = objekt;
	}

	/**
	 * {@inheritDoc}.<br>
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.Kopierbar#erzeugeKopie()
	 */
	@Override
	public Kopierbar erzeugeKopie() {
		Kopierbar result = null;
		try {
			final Method methode = objekt.getClass().getDeclaredMethod("clone", new Class<?>[0]);
			result = new JavaObjekt(methode.invoke(objekt, new Object[0]));
		} catch (final Exception e) {
			result = new Fehler(e.getMessage());
		}

		return result;
	}

	/**
	 * führt eine Operation mit dem Java-Objekt aus.<br>
	 *
	 * @param kontext
	 *            der Kontext für die Ausführung
	 * @param name
	 *            der Name der auszuführenden Operation
	 * @param operanden
	 *            die Ausdrücke für die Beschreibung der zu übergebenden
	 *            Parameter
	 * @return das Ergebnis der Funktion
	 */
	public Object execute(final Kontext kontext, final String name, final List<? extends Ausdruck> operanden) {
		final Object[] argumente = new Object[operanden.size()];
		int idx = 0;
		for (final Ausdruck ausdruck : operanden) {
			argumente[idx] = ausdruck.interpret(kontext);
			idx++;
		}

		final Class<?>[] types = new Class[operanden.size()];
		for (idx = 0; idx < operanden.size(); idx++) {
			Class<?> cl = argumente[idx].getClass();
			if (cl.equals(Long.class)) {
				cl = long.class;
			}
			types[idx] = cl;
		}

		Object result = null;
		try {
			Method methode = null;
			Method methodeProxy = null;

			try {
				methode = objekt.getClass().getDeclaredMethod(name, types);
			} catch (final NoSuchMethodException e) {

				final Method[] methods = objekt.getClass().getMethods();

				for (final Method method : methods) {
					switch (match(method, name, argumente)) {
					case FULL:
						methode = method;
						break;
					case NEAR:
						methodeProxy = method;
						break;
					case NONE:
						break;
					default:
						JavaObjekt.LOGGER.warning(e.getLocalizedMessage());
						break;
					}

					if (methode != null) {
						break;
					}
				}
			}

			if ((methode == null) && (methodeProxy != null)) {
				konvertiereArgumente(methodeProxy, argumente);
				methode = methodeProxy;
			}

			if (methode != null) {
				final Object funcResult = methode.invoke(objekt, argumente);
				if (funcResult != null) {
					result = new JavaObjekt(funcResult);
				}
			} else {
				throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE, "Methode nicht gefunden: " + name);
			}

		} catch (final SecurityException e) {
			throw new InterpreterException(e);
		} catch (final IllegalArgumentException e) {
			throw new InterpreterException(e);
		} catch (final IllegalAccessException e) {
			throw new InterpreterException(e);
		} catch (final InvocationTargetException e) {
			throw new InterpreterException(e);
		}
		return result;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur#getStrukturElement(java.lang.String)
	 * @see #setStrukturElement(String, Object)
	 */
	@Override
	public Object getStrukturElement(final String name) {
		try {
			final Field attribut = objekt.getClass().getDeclaredField(name);
			return new JavaObjekt(attribut.get(objekt));
		} catch (final SecurityException e) {
			throw new InterpreterException(e);
		} catch (final IllegalArgumentException e) {
			throw new InterpreterException(e);
		} catch (final NoSuchFieldException e) {
			throw new InterpreterException(e);
		} catch (final IllegalAccessException e) {
			throw new InterpreterException(e);
		}
	}

	/**
	 * konvertiert die übergebenen Argumente, so, dass sie für die übergebene
	 * Methode anwendbar sind.
	 *
	 * @param method
	 *            die Methode, auf die die Anpassung erfolgen soll
	 * @param argumente
	 *            die Argumente für die Ausführung der Methode
	 */
	private void konvertiereArgumente(final Method method, final Object[] argumente) {

		final Class<?>[] types = method.getParameterTypes();

		for (int idx = 0; idx < types.length; idx++) {
			if (!types[idx].isInstance(argumente[idx])) {
				argumente[idx] = JavaHandler.konvertiere(types[idx], argumente[idx]);
			}
		}
	}

	/**
	 * ermittelt, ob die übergebende Methode mit den übergebenen Parametern
	 * ausgeführt werden kann.
	 *
	 * @param method
	 *            die Methode, die geprüft werden soll.
	 * @param name
	 *            der Name der Methode
	 * @param argumente
	 *            die Argumente
	 * @return die Passgenauigkeit
	 */
	private MatchingQuality match(final Method method, final String name, final Object[] argumente) {
		MatchingQuality result = MatchingQuality.NONE;

		if (method.getName().equals(name)) {
			final Class<?>[] types = method.getParameterTypes();
			if (types.length == argumente.length) {
				if (types.length == 0) {
					result = MatchingQuality.FULL;
				} else {
					int matchCount = 0;
					int konvertCount = 0;
					for (int idx = 0; idx < types.length; idx++) {
						if (types[idx].isInstance(argumente[idx])) {
							matchCount++;
						} else if (JavaHandler.konvertierbar(types[idx], argumente[idx])) {
							konvertCount++;
						} else {
							break;
						}
					}
					if (matchCount == types.length) {
						result = MatchingQuality.FULL;
					} else if ((matchCount + konvertCount) == types.length) {
						result = MatchingQuality.NEAR;
					}
				}
			}
		}

		return result;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur#setStrukturElement(java.lang.String,
	 *      java.lang.Object)
	 * @see #getStrukturElement(String)
	 */
	@Override
	public void setStrukturElement(final String name, final Object daten) {
		try {
			final Field attribut = objekt.getClass().getDeclaredField(name);
			attribut.set(objekt, daten);
		} catch (final SecurityException e) {
			throw new InterpreterException(e);
		} catch (final IllegalArgumentException e) {
			throw new InterpreterException(e);
		} catch (final NoSuchFieldException e) {
			throw new InterpreterException(e);
		} catch (final IllegalAccessException e) {
			throw new InterpreterException(e);
		}
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		if (objekt != null) {
			return objekt.toString();
		}
		return super.toString();
	}
}
