/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.daten.zeit;

import de.bsvrz.ibv.uda.interpreter.Kopierbar;
import de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ObjektZugriffsFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischerWert;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation eines Zeitbereichs für Archivabfragen innerhalb eines
 * Uda-Skripts.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class ZeitBereich implements Struktur, Kopierbar {

	/**
	 * der Anfangszeitpunkt.
	 */
	private Zeitstempel start;

	/**
	 * der Endzeitpunkt.
	 */
	private Zeitstempel ende;

	/**
	 * der Endzeitpunkt gehört mit zum Bereich ?
	 */
	private boolean inklusive;

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt ein Zeitbereichsobjekt mit den angegebenen
	 * Grenzwerten.
	 *
	 * @param start
	 *            der Zeitstempel des Anfangzeitpunkts
	 * @param ende
	 *            der Zeitstempel des Endezeitpunkts oder die Zeitdauer als
	 *            Differenz zum Anfangszeitpunkt.
	 * @param inklusive
	 *            true, wenn der Endzeitpunkt selbst mit in das Intervall gehört
	 */
	public ZeitBereich(final Zeitstempel start, final Object ende, final boolean inklusive) {
		this.start = start;
		if (ende instanceof Zeitstempel) {
			this.ende = (Zeitstempel) ende;
		} else if (ende instanceof ZeitDauer) {
			this.ende = start;
			this.ende.add(((ZeitDauer) ende).getMilliSekunden());
		} else {
			throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE);
		}
		this.inklusive = inklusive;
	}

	/**
	 * {@inheritDoc}.<br>
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.Kopierbar#erzeugeKopie()
	 */
	@Override
	public ZeitBereich erzeugeKopie() {
		return new ZeitBereich(start, ende, inklusive);
	}

	/**
	 * liefert den Endzeitpunkt des Zeitbereichs in Millisekunden.
	 *
	 * @return den Zeitpunkt
	 */
	public Zeitstempel getEnde() {
		return ende;
	}

	/**
	 * liefert den Anfangszeitpunkt des Zeitbereichs in Millisekunden.
	 *
	 * @return den Zeitpunkt
	 */
	public Zeitstempel getStart() {
		return start;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Gültige Attribute sind:
	 * <ul>
	 * <li>anfangsZeit: für die Anfangszeit des Bereichs als Zeitstempel</li>
	 * <li>endZeit: für die Endezeit des Bereichs als Zeitstempel</li>
	 * <li>inklusive: liefert einen LogsichenWert, der beschreibt, ob der
	 * Endzeitpunkt selbst zum Bereich gehört</li>
	 * </ul>
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur#getStrukturElement(java.lang.String)
	 */
	@Override
	public Object getStrukturElement(final String name) {
		Object result = null;
		if ("anfangsZeit".equals(name)) {
			result = start;
		} else if ("endZeit".equals(name)) {
			result = ende;
		} else if ("inklusive".equals(name)) {
			result = LogischerWert.of(inklusive);
		} else {
			throw new ObjektZugriffsFehler(UdaFehlerSubtyp.ATTRIBUT, "Ungültiges Attribut: " + name);
		}
		return result;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.ausdruck.Struktur#setStrukturElement(java.lang.String,
	 *      java.lang.Object)
	 */
	@Override
	public void setStrukturElement(final String name, final Object daten) {
		if (daten == null) {
			throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE, "Leere Daten für Attribut: " + name);
		}
		try {
			if ("anfangsZeit".equals(name)) {
				start = (Zeitstempel) daten;
			} else if ("endZeit".equals(name)) {
				ende = (Zeitstempel) daten;
			} else if ("inklusive".equals(name)) {
				inklusive = ((LogischerWert) daten).get();
			} else {
				throw new ObjektZugriffsFehler(UdaFehlerSubtyp.ATTRIBUT, "Ungültiges Attribut: " + name);
			}
		} catch (final ClassCastException e) {
			final String message = "Ungültiger Datentyp: " + daten.getClass().getName()
					+ " für Attribut: " + name;
			Debug.getLogger().error(message, e);
			throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE, message);
		}
	}

}
