/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.verwaltung.protokoll;

import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import com.bitctrl.Constants;

import de.bsvrz.ibv.uda.uda.data.SkriptLauf;
import de.bsvrz.ibv.uda.verwaltung.ServerSkriptLauf;

/**
 * beschreibt die Protokollierung der Ausgabedaten eines Uda-Skripts in einen
 * Datei-Ringpuffer (gemäß Java-Standard-Logging).
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class LogProtokoll extends Protokoll {

	/**
	 * der Formatierer für die Ausgaben des Protokolls.<br>
	 * Die Daten werden in der Form &lt;Zeitstempel&gt;: &lt;Ausgabetext&gt; in
	 * die Logdateien ausgegeben.<br>
	 * Dieser Typ eines Protokoll wird nur von Skriptläufen unterstützt, die auf
	 * einem Server "Umfassende datenanalyse" ausgeführt werden.
	 *
	 * @author peuker
	 */
	public class LogProtokollFormatter extends Formatter {

		/**
		 * {@inheritDoc}.<br>
		 * Die Funktion erzeugt den Logeintrag in der Form &lt;Zeitstempel&gt;:
		 * &lt;Ausgabetext&gt;.
		 *
		 * @see java.util.logging.Formatter#format(java.util.logging.LogRecord)
		 */
		@Override
		public String format(final LogRecord record) {
			return DateFormat.getDateTimeInstance().format(new Date(System.currentTimeMillis())) + ": "
					+ record.getMessage() + Constants.NL;
		}
	}

	/**
	 * die Versions-ID für die Serialisierung.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * der Logger, der für die Ausgabe der Daten verwendet werden soll.
	 */
	private transient Logger logger;

	/**
	 * Standardkonstruktor.
	 */
	public LogProtokoll() {
		super(ProtokollTyp.LOGGING);
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion schliesst die Handler des Loggers und entfernt den Logger
	 * selbst.
	 *
	 * @see de.bsvrz.ibv.uda.verwaltung.protokoll.Protokoll#abschliessen()
	 */
	@Override
	public void abschliessen() {
		if (logger != null) {
			for (final Handler handler : logger.getHandlers()) {
				handler.flush();
				handler.close();
				logger.removeHandler(handler);
			}
			logger = null;
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion gibt die Daten der Skriptausgabe als INFO-Meldung über den
	 * Logger aus. Die Formatierung der Daten erfolgt mit dem
	 * {@link LogProtokollFormatter}.
	 *
	 * @see de.bsvrz.ibv.uda.verwaltung.protokoll.Protokoll#ausgeben(de.bsvrz.ibv.uda.verwaltung.protokoll.ProtokollEintrag)
	 */
	@Override
	public void ausgeben(final ProtokollEintrag text) {
		if (logger != null) {
			logger.info(text.getText());
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion initialisiert den Logger für die Ausgaben des Skriptlaufs.
	 * Es wird ein Filehandler installert, der den für das Protokoll definierten
	 * Namen für die Logdatei, die Anzahl der Logdateien und die maximale Größe
	 * für die Logdateien übergeben bekommt.<br>
	 * Die Logdateien werden immer in das Verzeichnis &lt;UdaModul-PID&gt;/log
	 * ausgegeben. Wurde kein Name definiert, ergibt sich der Name aus
	 * &lt;Skriptlauf-PID&gt;.log.<br>
	 * Die Anzahl für die Logdateien muss mindestens 1 betragen und wird bei
	 * Bedarf automatisch korrigiert.<br>
	 * Die maximale Größe einer Logdatei muss größer-gleich 0 sein und wird
	 * ebenfalls automatsich korrigiert. Der Wert für die Größe steht hierbei
	 * für eine unbegrenzte Größe.
	 *
	 * @see de.bsvrz.ibv.uda.verwaltung.protokoll.Protokoll#initialisiere(de.bsvrz.ibv.uda.uda.data.SkriptLauf)
	 */
	@Override
	public void initialisiere(final SkriptLauf lauf) {
		logger = null;
		if (lauf instanceof ServerSkriptLauf) {
			final ServerSkriptLauf serverLauf = (ServerSkriptLauf) lauf;
			logger = Logger.getLogger(lauf.getName());
			for (final Handler handler : logger.getHandlers()) {
				logger.removeHandler(handler);
			}

			String name = getName();
			if ((name == null) || (name.length() <= 0)) {
				name = serverLauf.getObjekt().getPid() + ".log";
			}

			int anzahl = (int) getAnzahl();
			if (anzahl <= 0) {
				anzahl = 1;
			}

			int limit = (int) getMaxSize();
			if (limit < 0) {
				limit = 0;
			} else {
				limit *= (1024 * 1024);
			}

			final File file = new File(serverLauf.getModul().getModulVerzeichnis(), "log");
			file.mkdirs();

			final FileHandler handler;
			try {
				handler = new FileHandler(new File(file, name).getAbsolutePath(), limit, anzahl);
				handler.setFormatter(new LogProtokollFormatter());
				logger.addHandler(handler);
			} catch (final IOException e) {
				e.printStackTrace();
				logger = null;
			}
		}
	}
}
