/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.client;

import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPublicKeySpec;
import java.text.DateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.uda.interpreter.daten.dav.BestaetigterDavSender;
import de.bsvrz.ibv.uda.uda.UdaMessage;
import de.bsvrz.ibv.uda.uda.data.Modul;
import de.bsvrz.ibv.uda.uda.data.Skript;
import de.bsvrz.ibv.uda.uda.data.UDA;
import de.bsvrz.ibv.uda.verwaltung.protokoll.Protokoll;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;

/**
 * die Repräsentation eines UdaModuls in einem Client einer Uda-Applikation.<br>
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class ClientModul extends Modul implements MutableSetChangeListener, ClientReceiverInterface {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * das Objekt, das den Typ für Uda-Skripte innerhalb der
	 * Datenverteiler-Konfiguration definiert.
	 */
	private final DynamicObjectType skriptTyp;

	/**
	 * die Menge der Skript des Uda-Moduls innerhalb der
	 * Datenverteilerkonfiguration.
	 */
	private final MutableSet skriptMenge;

	/**
	 * die Liste der Skripte, die innerhalb des Moduls definiert sind.
	 */
	private final Map<Long, ClientSkript> skripte = new HashMap<>();

	/**
	 * das Objekt, das den Typ für Uda-Skriptläufe innerhalb der
	 * Datenverteiler-Konfiguration definiert.
	 */
	private final DynamicObjectType skriptLaufTyp;

	/**
	 * die Liste der Skriptläufe, die innerhalb des Moduls definiert sind.
	 */
	private final Map<Long, ClientSkriptLauf> laeufe = new HashMap<>();

	/**
	 * die Menge der Skriptläufe des Uda-Moduls innerhalb der
	 * Datenverteilerkonfiguration.
	 */
	private final MutableSet laufMenge;

	/**
	 * die Liste der Listener, die benachrichtigt werden, wenn sich in der Liste
	 * der Skripte bzw. der Skriptläfe Änderungen ergeben haben.
	 */
	private final Set<ModulListener> listenerListe = new HashSet<>();

	/**
	 * die Liste der Listener, die benachrichtigt werden, wenn sich die Daten
	 * eines des Skripte des Moduls ändern.
	 */
	private final Set<SkriptListener> skriptListenerListe = new HashSet<>();

	/**
	 * die Liste der Listener, die benachrichtigt werden, wenn sich die Daten
	 * eines der Skriptläufe des Moduls ändern.
	 */
	private final Set<SkriptLaufListener> laufListenerListe = new HashSet<>();

	/**
	 * die Datenbeschreibung für die Versendung des öffentlich Schlüssels der
	 * zur verschlüsselung des Passworts für die Übertragung verwendet wird.
	 */
	private final DataDescription keyDesc;

	/**
	 * der öffentliche Schlüssel für die Passwortverschlüsselung.
	 */
	private RSAPublicKey publicKey;

	/**
	 * der Sender zur Übertragung von Daten zum Datenverteiler.
	 */
	private final BestaetigterDavSender sender;

	/**
	 * Konstruktor.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @param modulObjekt
	 *            das Systemobjekt, welches das Uda-Modul innerhalb der
	 *            Datenverteiler-Konfiguration repräsentiert
	 */
	ClientModul(final ClientDavInterface verbindung, final ConfigurationObject modulObjekt) {
		super(verbindung, modulObjekt);

		sender = BestaetigterDavSender.getSender(verbindung);

		keyDesc = new DataDescription(verbindung.getDataModel().getAttributeGroup(UDA.MODUL_ATG_KEY),
				verbindung.getDataModel().getAspect(UDA.ASP_IST));
		verbindung.subscribeReceiver(this, modulObjekt, keyDesc, ReceiveOptions.normal(), ReceiverRole.receiver());

		skriptTyp = (DynamicObjectType) verbindung.getDataModel().getType(UDA.TYP_SKRIPT);
		skriptLaufTyp = (DynamicObjectType) verbindung.getDataModel().getType(UDA.TYP_LAUF);

		/*
		 * Ermitteln der Menge aller Skripte und Einlesen der aktuell
		 * enthaltenen Objekte
		 */
		skriptMenge = modulObjekt.getMutableSet(UDA.MODUL_MENGE_SKRIPTE);
		update(skriptMenge, skriptMenge.getElements().toArray(new SystemObject[0]), new SystemObject[0]);
		skriptMenge.addChangeListener(this);

		/*
		 * Ermitteln der Menge aller Skriptläufe und Einlesen der aktuell
		 * enthaltenen Objekte
		 */
		laufMenge = modulObjekt.getMutableSet(UDA.MODUL_MENGE_LAEUFE);
		update(laufMenge, laufMenge.getElements().toArray(new SystemObject[0]), new SystemObject[0]);
		laufMenge.addChangeListener(this);
	}

	/**
	 * einen {@link ModulListener} hinzufügen, der bei Änderungen der Mengen der
	 * Skripte bzw. Skriptläufe benachrichtigt wird.
	 *
	 * @param listener
	 *            der neue Listener
	 * @see #removeClientListener(ModulListener)
	 */
	public void addClientListener(final ModulListener listener) {
		synchronized (listenerListe) {
			if (!listenerListe.contains(listener)) {
				listenerListe.add(listener);
			}
		}
	}

	/**
	 * fügt der Menge der Listener, die bei einer Änderung des Skriptlaufs
	 * benachrichtigt werden den übergebenen hinzu.
	 *
	 * @param listener
	 *            der neue Listener
	 * @see #removeLaufListener(SkriptLaufListener)
	 */
	public void addLaufListener(final SkriptLaufListener listener) {
		synchronized (laufListenerListe) {
			laufListenerListe.add(listener);
		}
	}

	/**
	 * fügt der Menge der Listener, die bei einer Änderung des Skripts
	 * benachrichtigt werden den übergebenen hinzu.
	 *
	 * @param listener
	 *            der neue Listener
	 * @see #removeSkriptListener(SkriptListener)
	 */
	public void addSkriptListener(final SkriptListener listener) {
		synchronized (skriptListenerListe) {
			skriptListenerListe.add(listener);
		}
	}

	/**
	 * aktualisiert die Liste der Skriptläufe innerhalb des UdaModuls.
	 *
	 * @param added
	 *            die der Menge hinzugefügten Objekte
	 * @param removed
	 *            die aus Menge entfernten Objekte
	 */
	private void aktualisiereSkriptLaufListe(final SystemObject[] added, final SystemObject[] removed) {

		synchronized (laeufe) {
			for (final SystemObject obj : added) {
				final ClientSkriptLauf lauf = new ClientSkriptLauf(this, obj);
				laeufe.put(obj.getId(), lauf);
				synchronized (listenerListe) {
					for (final ModulListener listener : listenerListe) {
						listener.skriptLaufAngelegt(this, lauf);
					}
				}
			}
			for (final SystemObject obj : removed) {
				final ClientSkriptLauf lauf = laeufe.get(obj.getId());
				if (lauf != null) {
					laeufe.remove(obj.getId());
					lauf.remove();
					synchronized (listenerListe) {
						for (final ModulListener listener : listenerListe) {
							listener.skriptLaufEntfernt(this, lauf);
						}
					}
				}
			}
		}
	}

	/**
	 * aktualisiert die Liste der Skripte innerhalb des UdaModuls.
	 *
	 * @param added
	 *            die der Menge hinzugefügten Objekte
	 * @param removed
	 *            die aus Menge entfernten Objekte
	 */
	private void aktualisiereSkriptListe(final SystemObject[] added, final SystemObject[] removed) {

		synchronized (skripte) {
			for (final SystemObject obj : added) {
				final ClientSkript skript = new ClientSkript(this, obj);
				skripte.put(obj.getId(), skript);
				synchronized (listenerListe) {
					for (final ModulListener listener : listenerListe) {
						listener.skriptAngelegt(this, skript);
					}
				}
			}
			for (final SystemObject obj : removed) {
				final ClientSkript skript = skripte.get(obj.getId());
				if (skript != null) {
					skripte.remove(obj.getId());
					skript.remove();
					synchronized (listenerListe) {
						for (final ModulListener listener : listenerListe) {
							listener.skriptEntfernt(this, skript);
						}
					}
				}
			}
		}
	}

	/**
	 * leert die intern verwalteten Listen von Listenern.
	 */
	void clear() {
		listenerListe.clear();
		skriptListenerListe.clear();
		laufListenerListe.clear();
	}

	/**
	 * die Funktion verschlüsselt eine Zeichenkette unter Verwendung des von der
	 * "Umfassenden Datenanalyse" zur Verfügung gestellten öffentlichen
	 * Schlüssels.
	 *
	 * @param text
	 *            der zu verschlüsselnde Text
	 * @return die verschlüsselte Bytefolge
	 * @throws ClientException
	 *             die Verschlüsselung konnte nicht ausgeführt werden
	 */
	byte[] encrypt(final String text) throws ClientException {
		try {
			final Cipher cipher = Cipher.getInstance("RSA");
			cipher.init(Cipher.ENCRYPT_MODE, publicKey);
			return cipher.doFinal(text.getBytes());
		} catch (final InvalidKeyException | IllegalBlockSizeException | BadPaddingException | NoSuchAlgorithmException
				| NoSuchPaddingException e) {
			ClientModul.LOGGER.info(e.getMessage());
			throw new ClientException("Passwortverschlüsselung nicht möglich", e);
		}
	}

	/**
	 * die Funktion erzeugt ein neues Skriptobjekt für das Uda-Modul innerhalb
	 * der Datenverteiler-Konfiguration.<br>
	 * Es wird ein neues dynamisches Objekt des entsprechenden Typs angelegt.
	 * Das Objekt wird in die Menge des UdaModuls eingetragen. Nach der
	 * Rückmeldung der Mengenänderung und der Erzeugung eines internen
	 * Skriptobjektes, wird diese mit den Daten des Nutzers, dem Skriptnamen und
	 * einem initialen Quelltext versorgt.
	 *
	 * @param name
	 *            der Name des Skripts
	 * @param username
	 *            der Name des Nutzers der das Skript anlegt
	 * @return das Skript das Objekt, das das angelegte Skript innerhalb des
	 *         Clients repräsentiert.
	 * @throws ClientException
	 *             das Skript konnte nicht angelegt werden
	 */
	public ClientSkript erzeugeSkript(final String name, final String username) throws ClientException {

		final Long now = System.currentTimeMillis();
		try {
			final DynamicObject skriptObj = DynamischeObjekte.getInstanz(getVerbindung()).erzeugeObjekt(skriptTyp,
					"uda.skript." + now, "uda.skript." + now);
			getModulObjekt().getMutableSet(UDA.MODUL_MENGE_SKRIPTE).add(skriptObj);
			final ClientSkript skript = waitForAddedSkript(skriptObj, 20000L);
			if (skript != null) {
				skript.setSkriptName(name);
				skript.setErstelltAm(now);
				skript.setErstelltDurch(username);
				skript.addSourceLine("# Skript        : " + name);
				skript.addSourceLine("# Erstellt durch: " + username);
				skript.addSourceLine("# Am            : " + DateFormat.getDateTimeInstance().format(new Date(now)));
				skript.save(username, "Skript erstellt");
			}
		} catch (final ConfigurationChangeException | DynObjektException e) {
			throw new ClientException("Das Skript konnte nicht erstellt werden", e);
		}

		return null;
	}

	/**
	 * die Funktion erzeugt ein neues Skriptlaufobjekt für das Uda-Modul
	 * innerhalb der Datenverteiler-Konfiguration.<br>
	 * Es wird ein neues dynamisches Objekt des entsprechenden Typs angelegt.
	 * Das Objekt wird in die Menge des UdaModuls eingetragen. Nach der
	 * Rückmeldung der Mengenänderung und der Erzeugung eines internen
	 * Skriptobjektes, wird diese mit den entsprechenden Daten versorgt.
	 *
	 * @param daten
	 *            die Daten des anzulegenden Skriptlaufs
	 * @param benutzer
	 *            der Name des Benutzers, der eine Änderung vornimmt
	 * @param grund
	 *            der Grund für die Änderung
	 * @return das Skript das Objekt, das das angelegte Skript innerhalb des
	 *         Clients repräsentiert.
	 * @throws ClientException
	 *             der Skriptlauf konnte nicht angelegt werden
	 */
	public ClientSkriptLauf erzeugeSkriptLauf(final SkriptLaufDaten daten, final String benutzer, final String grund)
			throws ClientException {

		final Long now = System.currentTimeMillis();
		try {
			final DynamicObject skriptLaufObj = DynamischeObjekte.getInstanz(getVerbindung())
					.erzeugeObjekt(skriptLaufTyp, "uda.skriptlauf." + now, "uda.skriptlauf." + now);

			getModulObjekt().getMutableSet(UDA.MODUL_MENGE_LAEUFE).add(skriptLaufObj);
			final ClientSkriptLauf skriptLauf = waitForAddedSkriptLauf(skriptLaufObj, 20000L);
			if (skriptLauf != null) {
				skriptLauf.setName(daten.getName());
				skriptLauf.setSkript(daten.getSkript());
				skriptLauf.setBenutzer(daten.getBenutzer());
				skriptLauf.setPasswd(daten.getPasswd());
				skriptLauf.setBeschreibung(daten.getBeschreibung());

				skriptLauf.clearParameter();
				for (final String key : daten.getSkriptParameter().keySet()) {
					skriptLauf.setParameter(key, daten.getSkriptParameter().get(key));
				}

				skriptLauf.setStartTyp(daten.getStartTyp());
				skriptLauf.setZyklus(daten.getZyklus());

				skriptLauf.clearProtokolle();
				for (final Protokoll protokoll : daten.getProtokolle()) {
					skriptLauf.addProtokoll(protokoll);
				}

				String grundText = grund;
				if (grundText == null) {
					grundText = "Skriptlauf angelegt";
				}
				skriptLauf.save(benutzer, grundText);
			}
		} catch (final ConfigurationChangeException | DynObjektException e) {
			throw new ClientException("Der Skriptlauf konnte nicht erstellt werden", e);
		}

		return null;
	}

	/**
	 * liefert die Liste der Skriptlauflistener.
	 *
	 * @return laufListenerListe
	 */
	Set<SkriptLaufListener> getLaufListenerListe() {
		return laufListenerListe;
	}

	/**
	 * liefert das SenderObjekt, über das alle Sendeoperationen des Clients die
	 * mit diesem Modul in Verbindung stehen ausgeführt werden.
	 *
	 * @return den Sender
	 */
	BestaetigterDavSender getSender() {
		return sender;
	}

	/**
	 * ermittelt das Skript innerhalb des Moduls, das durch das Systemobjekt mit
	 * der übergebenen PID repräsentiert wird.
	 *
	 * @param pid
	 *            die PID des Systemobjekts
	 * @return das ermittelte Skript oder <i>null</i>, wenn kein solches
	 *         existiert
	 */
	@Override
	public Skript getSkript(final String pid) {
		Skript result = null;
		for (final ClientSkript skript : skripte.values()) {
			if ((skript.getDavObjekt() != null) && (skript.getDavObjekt().getPid().equals(pid))) {
				result = skript;
				break;
			}
		}
		return result;
	}

	/**
	 * liefert die Repräsentation der Menge aller Skriptläufe des Moduls zurück.
	 *
	 * @return die Liste der Skriptläufe
	 */
	public Collection<ClientSkriptLauf> getSkriptLaeufe() {
		synchronized (laeufe) {
			return laeufe.values();
		}
	}

	/**
	 * liefert die Liste der Skriptlistener.
	 *
	 * @return skriptListenerListe
	 */
	Set<SkriptListener> getSkriptListenerListe() {
		return skriptListenerListe;
	}

	/**
	 * liefert die Repräsentation der Menge aller Skripte des Moduls zurück.
	 *
	 * @return die Liste der Skripte
	 */
	public ClientSkript[] getSkripts() {
		synchronized (skripte) {
			return skripte.values().toArray(new ClientSkript[skripte.size()]);
		}
	}

	/**
	 * ermittelt das Skript mit dem angegebenen Namen oder <i>null</i>, wenn
	 * keines gefunden wurde.
	 *
	 * @param name
	 *            der Name des gesuchten Skripts
	 * @return das Skript oder <i>null</i>
	 */
	public Skript getSkriptWithName(final String name) {
		Skript result = null;
		for (final ClientSkript skript : skripte.values()) {
			if (skript.getName().equals(name)) {
				result = skript;
				break;
			}
		}
		return result;
	}

	/**
	 * ermittelt, ob das Modul gültig ist.<br>
	 * Das UdaModul-Objekt wird als gültig betrachtet, wenn eine Anwendung vom
	 * Typ "Umfassende Datenanalyse" für dieses Objekt angemeldet ist.
	 *
	 * @return <i>true</i>, wenn das Modul gültig ist
	 */
	public boolean isValid() {
		return publicKey != null;
	}

	/**
	 * liefert des Skript mit dem angegebenen Namen.<br>
	 * Da Namen für Skripte nicht eindeutig sind, wird jeweils das zuerst
	 * gefundene geliefert. Welches Skript bei mehreren mit gleichem Namen
	 * zuerst gefunden wird ist nicht deinfiert. Existiert kein Skript mit dem
	 * gesuchten Namen, wird der Wert <i>null</i> zurückgeliefert.
	 *
	 * @param name
	 *            der Name der Name des gesuchten Skripts
	 * @return das Skript oder <i>null</i>, wenn keines existiert
	 */
	public ClientSkript loadSkript(final String name) {
		ClientSkript result = null;
		for (final ClientSkript skript : getSkripts()) {
			if (name.equals(skript.getName())) {
				result = skript;
				break;
			}
		}
		return result;
	}

	/**
	 * einen {@link ModulListener} vom Modul abmelden.<br>
	 * Der übergebene Listener muss nicht in der Menge enthalten sein. Dieser
	 * Fall wird ignoriert.
	 *
	 * @param listener
	 *            der abzumeldende Listener
	 * @see #addClientListener(ModulListener)
	 */
	public void removeClientListener(final ModulListener listener) {
		synchronized (listenerListe) {
			listenerListe.remove(listener);
		}
	}

	/**
	 * entfernt den angegebenen Skriptlauf.
	 *
	 * @param lauf
	 *            der Skriptlauf
	 * @throws ConfigurationChangeException
	 *             der Lauf konnte nicht entfernt werden
	 */
	public void removeLauf(final ClientSkriptLauf lauf) throws ConfigurationChangeException {
		try {
			lauf.beenden();
		} catch (final ClientException e) {
			ClientModul.LOGGER.warning(e.getLocalizedMessage());
			// der Fehler kann hier ignoriert werden, da der Lauf eventuell
			// nicht aktiv wahr
		}

		laufMenge.remove(lauf.getDavObjekt());
	}

	/**
	 * entfernt einen Listener für den Skriptlauf.
	 *
	 * @param listener
	 *            der zu entfernende Listener
	 * @see #addLaufListener(SkriptLaufListener)
	 */
	public void removeLaufListener(final SkriptLaufListener listener) {
		synchronized (laufListenerListe) {
			laufListenerListe.remove(listener);
		}
	}

	/**
	 * entfernt das Skript vom Modul.<br>
	 * Das Skript wird aus der Menge der Skripte innerhalb der
	 * Datenverteilerkonfiguration entfernt und das Objekt wird als ungültig
	 * erklärt. Die Entfernung des Skripts aus der internen Liste des Moduls
	 * innerhalb des Clients erfolgt über die Rückmeldung der Mengenänderung.
	 *
	 * @param skript
	 *            das zu entfernende Skript
	 * @throws ConfigurationChangeException
	 *             beim Entfernen des Sktipts ist ein Fehler aufgetreten
	 */
	public void removeSkript(final ClientSkript skript) throws ConfigurationChangeException {
		for (final ClientSkriptLauf lauf : getSkriptLaeufe()) {
			if (skript.equals(lauf.getSkript())) {
				try {
					lauf.beenden();
				} catch (final ClientException e) {
					ClientModul.LOGGER.warning(e.getLocalizedMessage());
					// der Fehler kann ignoriert werden
				}
				lauf.setSkript(null);
				try {
					lauf.save("System", "Skript entfernt");
				} catch (final ClientException e) {
					ClientModul.LOGGER.warning(e.getLocalizedMessage());
					// der Fehler kann hier ignoriert werden, da der Skriptlauf
					// ohnehin entfernt wird
				}
			}
		}

		skriptMenge.remove(skript.getDavObjekt());
	}

	/**
	 * entfernt einen Listener für das Skript.
	 *
	 * @param listener
	 *            der zu entfernende Listener
	 * @see #addSkriptListener(SkriptListener)
	 */
	public void removeSkriptListener(final SkriptListener listener) {
		synchronized (skriptListenerListe) {
			skriptListenerListe.remove(listener);
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion erwartet neue oder gelöschte Einträge innerhalb der Liste
	 * der Skripte bzw. der Skriptläufe und legt entsprechende interne Objekte
	 * zur Repräsentation an bzw. entfernt diese. Die Übertragung der Daten
	 * erfolgt über die entsprechenden Attributgruppen der Objekte selbst.
	 *
	 * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet,
	 *      de.bsvrz.dav.daf.main.config.SystemObject[],
	 *      de.bsvrz.dav.daf.main.config.SystemObject[])
	 */
	@Override
	public void update(final MutableSet set, final SystemObject[] added, final SystemObject[] removed) {
		if (set.equals(skriptMenge)) {
			aktualisiereSkriptListe(added, removed);
		} else if (set.equals(laufMenge)) {
			aktualisiereSkriptLaufListe(added, removed);
		} else {
			LogTools.log(ClientModul.LOGGER, UdaMessage.WarningUnerwartetesMengenUpdate, set);
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion aktualisiert den von der "Umfassenden Datenanalyse"
	 * bereitgestellten öffentlichen Schlüssel.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public void update(final ResultData[] results) {
		final RSAPublicKey oldKey = publicKey;

		for (final ResultData result : results) {
			if (result.getDataDescription().getAttributeGroup().equals(keyDesc.getAttributeGroup())) {
				if (result.getDataDescription().getAspect().equals(keyDesc.getAspect())) {
					if (result.hasData()) {
						final Data data = result.getData();
						Data.NumberArray array = data.getArray("Modulus").asUnscaledArray();
						final byte[] modulus = new byte[array.getLength()];
						for (int idx = 0; idx < array.getLength(); idx++) {
							modulus[idx] = array.getValue(idx).byteValue();
						}

						array = data.getArray("Exponent").asUnscaledArray();
						final byte[] exponent = new byte[array.getLength()];
						for (int idx = 0; idx < array.getLength(); idx++) {
							exponent[idx] = array.getValue(idx).byteValue();
						}

						try {
							publicKey = (RSAPublicKey) KeyFactory.getInstance("RSA").generatePublic(
									new RSAPublicKeySpec(new BigInteger(modulus), new BigInteger(exponent)));
						} catch (final InvalidKeySpecException e) {
							ClientModul.LOGGER.info(e.getMessage());
						} catch (final NoSuchAlgorithmException e) {
							ClientModul.LOGGER.info(e.getMessage());
						}

					} else {
						publicKey = null;
					}
				}
			}
		}

		if (oldKey != publicKey) {
			synchronized (listenerListe) {
				for (final ModulListener listener : listenerListe) {
					listener.neuerModulStatus(this, isValid());
				}
			}
		}
	}

	/**
	 * wartet, ob ein Skript angelegt wurde.<br>
	 * Die Funktion wird intern beim Anlegen eines Skripts verwendet. Sie läuft
	 * solange in einer Schleife, bis über die Datenverteilerrückmeldung das
	 * gewünschte Objekt innhalb der Skriptmenge eingetragen wurde. Zusätzlich
	 * kann eine maximale Wartezeit auf das Ergebnis in Miliisekunden angegeben
	 * werden. Ist die Wartezeit abgelaufen bevor das Element in der Menge
	 * eingetragen wurde, wird eine entsprechende Ausnahme erzeugt.
	 *
	 * @param skriptObj
	 *            das Objekt, das in der Menge erwartet wird.
	 * @param timeout
	 *            die maximale Wartezeit in Millisekunden
	 * @return das Skriptobjekt
	 * @throws ClientException
	 *             innerhalb der maximalen Wartezeit konnte kein entsprechendes
	 *             Skriptobjekt ermittelt werden.
	 */
	private synchronized ClientSkript waitForAddedSkript(final SystemObject skriptObj, final long timeout)
			throws ClientException {
		ClientSkript skript = null;
		long timeoutValue = Long.MAX_VALUE;
		if (timeout > 0) {
			timeoutValue = System.currentTimeMillis() + timeout;
		}
		do {
			if (System.currentTimeMillis() > timeoutValue) {
				throw new ClientException("Das Skript wurde nicht in der geforderten Zeit angelegt");
			}
			skript = skripte.get(skriptObj.getId());
			if ((skript == null) || !skript.isValid()) {
				getVerbindung().sleep(200);
			}
		} while ((skript == null) || !skript.isValid());

		return skript;
	}

	/**
	 * wartet, ob ein Skriptlauf angelegt wurde.<br>
	 * Die Funktion wird intern beim Anlegen eines Skriptlaufs verwendet. Sie
	 * läuft solange in einer Schleife, bis über die Datenverteilerrückmeldung
	 * das gewünschte Objekt innhalb der Skriptlaufmenge eingetragen wurde.
	 * Zusätzlich kann eine maximale Wartezeit auf das Ergebnis in Miliisekunden
	 * angegeben werden. Ist die Wartezeit abgelaufen bevor das Element in der
	 * Menge eingetragen wurde, wird eine entsprechende Ausnahme erzeugt.
	 *
	 * @param skriptObj
	 *            das Objekt, das in der Menge erwartet wird.
	 * @param timeout
	 *            die maximale Wartezeit in Millisekunden
	 * @return das Skriptobjekt
	 * @throws ClientException
	 *             innerhalb der maximalen Wartezeit konnte kein entsprechendes
	 *             Skriptobjekt ermittelt werden.
	 */
	private synchronized ClientSkriptLauf waitForAddedSkriptLauf(final SystemObject skriptObj, final long timeout)
			throws ClientException {
		ClientSkriptLauf lauf = null;
		long timeoutValue = Long.MAX_VALUE;
		if (timeout > 0) {
			timeoutValue = System.currentTimeMillis() + timeout;
		}
		do {
			if (System.currentTimeMillis() > timeoutValue) {
				throw new ClientException("Der Skriptlauf wurde nicht in der geforderten Zeit angelegt");
			}
			lauf = laeufe.get(skriptObj.getId());
			if ((lauf == null) || !lauf.isValid()) {
				getVerbindung().sleep(200);
			}
		} while ((lauf == null) || !lauf.isValid());

		return lauf;
	}

}
