/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.TimerTask;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.uda.interpreter.AnweisungsBlock;
import de.bsvrz.ibv.uda.interpreter.anweisung.RueckgabeAnweisung;
import de.bsvrz.ibv.uda.interpreter.anweisung.UdaAktivitaet;
import de.bsvrz.ibv.uda.interpreter.daten.SkriptKontext;
import de.bsvrz.ibv.uda.interpreter.daten.dav.DavDaten;
import de.bsvrz.ibv.uda.interpreter.daten.dav.MengenObjekt;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.DatenZugriffsFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.Fehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.KonfigurationsFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;

/**
 * die Repräsentation einer nebenläufigen Aktivität zum Empfang und Versenden
 * von Daten vom und zu einem verbundenen Datenverteiler.<br>
 * Die Klasse implementiert die Schnittstellen ClientReceiverInterface,
 * ClientSenderInterface und MutableSetChangeListener um Daten zum
 * Datenverteiler versenden zu können und Daten, sowie Änderungen von
 * dynamischen Mengen asynchron vom Datenverteiler empfangen und verrbeiten zu
 * können. Die Aktivität selbst realisiert keinen eigenen Thread. Die
 * Nebenläufigkeit wird durch die Kommunikationsthreads der
 * Datenverteiler-Applikationsfuntionen realisiert.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class Anmeldung extends UdaAktivitaet
implements MutableSetChangeListener, ClientReceiverInterface, ClientSenderInterface {

	/**
	 * die Klasse realisiert eine timergesteuerte Aufgabe, die zur Überwachung
	 * von Timeouts bei Empfängeranmeldungen sient.
	 *
	 * @author peuker
	 */
	private static class TimeoutChecker extends TimerTask {

		/**
		 * die Anmeldung, deren Emfängeranmeldungen überwacht werden sollen.
		 */
		private final Anmeldung anmeldung;

		/**
		 * Konstruktor.
		 *
		 * @param anmeldung
		 *            die zu überwachende Anmeldung
		 */
		TimeoutChecker(final Anmeldung anmeldung) {
			this.anmeldung = anmeldung;
		}

		/**
		 * {@inheritDoc}.<br>
		 * Die Funktion ermittelt für alle Empfängeranmeldung der überwachten
		 * Anmeldungsaktivität, ob diese in einen Timeoutstatus gelaufen sind.
		 * Ist dies der Fall wird der Fehlerbehandlungsblock ausgeführt.
		 *
		 * @see java.util.TimerTask#run()
		 */
		@Override
		public void run() {
			final long testZeit = System.currentTimeMillis();
			for (final AnmeldungsEintrag eintrag : anmeldung.getEmpfaengerAnmeldungen()) {
				if (eintrag instanceof AtgAnmeldungsEintrag) {
					final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
					if (atgAnmeldung.isTimedOut(testZeit)) {
						atgAnmeldung.setAktualisierung(testZeit);
						anmeldung.behandleFehler(new Fehler("Timeout: "
								+ atgAnmeldung));
					}
				}
			}
		}
	}

	/**
	 * die Funktion liefert ein Datenverteiler-Objekt mit dem übergebenen Typ.
	 * <br>
	 * Folgende Fälle sind möglich:
	 * <ul>
	 * <li>Das übergebene Objekt ist bereits eine Instanz des übergebenen Typs.
	 * </li>
	 * <li>Das übergebene Objekt ist eine internes Objekt, das ein
	 * Datenverteiler-Objekt des übergebenen Typs kapselt.</li>
	 * <li>Das übergebene Objekt ist eine Zeichenkette und das Objekt mit der
	 * der gleichnamigen PID entspricht dem übergebenen Typ.</li>
	 * </ul>
	 * In allen anderen Fällen wird ein Fehler ausgelöst.
	 *
	 * @param dav
	 *            die Datenverteilerverbindung
	 * @param obj
	 *            das zu prüfende Objekt
	 * @param type
	 *            der Typ des gewünschten Zielobjekts
	 * @return das Datenverteilerobjekt, das durch das übergebene Objekt
	 *         repräsentiert wird.
	 */
	static Object getDavObject(final ClientDavInterface dav, final Object obj, final Class<? extends Object> type) {
		Object result = null;

		if (type.isInstance(obj)) {
			result = obj;
		} else if (obj instanceof DavDaten) {
			result = ((DavDaten) obj).getObjekt();
		} else if (obj instanceof String) {
			result = dav.getDataModel().getObject((String) obj);
		}

		if ((result == null) || (!type.isInstance(result))) {
			throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE, "Abbildung von \"" + obj + "\" auf den Typ \""
					+ type.getSimpleName() + "\" nicht möglich");
		}
		return result;
	}

	/**
	 * die Anmeldeanweisung, auf der diese Aktivität beruht.
	 */
	private final AnmeldeAnweisung anweisung;

	/**
	 * die Datenverteilerverbindung.
	 */
	private final ClientDavInterface connection;

	/**
	 * die Liste der aktuellen Anmeldungen als Empfänger.
	 */
	private final Collection<AnmeldungsEintrag> empfaengerAnmeldungen = Collections
			.synchronizedCollection(new ArrayList<AnmeldungsEintrag>());

	/**
	 * die Liste der aktuellen Anmeldungen als Senke.
	 */
	private final Collection<AnmeldungsEintrag> senkeAnmeldungen = new ArrayList<>();

	/**
	 * die Liste der aktuellen Anmeldungen als Quelle.
	 */
	private final Collection<AnmeldungsEintrag> quelleAnmeldungen = new ArrayList<>();

	/**
	 * die Liste der aktuellen Anmeldungen als Sender.
	 */
	private final Collection<AnmeldungsEintrag> senderAnmeldungen = new ArrayList<>();

	/**
	 * Konstruktor.
	 *
	 * @param kontext
	 *            der umgebende Kontext der Anweisung
	 * @param name
	 *            der name unter dem die Aktivität im umgebenden Kontext
	 *            registriert wird
	 * @param anweisung
	 *            die Anweisung mit den Details für die Anmeldung beim
	 *            Datenverteiler und die auszuführenden Operationen
	 */
	public Anmeldung(final SkriptKontext kontext, final String name, final AnmeldeAnweisung anweisung) {
		super(kontext, name);
		this.anweisung = anweisung;
		connection = kontext.getInterpreter().getConnection();

		Object definitionsErgebnis = null;
		final AnweisungsBlock definitionsBlock = this.anweisung.getDefinitionAnweisungen();
		if (definitionsBlock != null) {
			definitionsErgebnis = definitionsBlock.interpret(getKontext());
		}

		if (!(definitionsErgebnis instanceof RueckgabeAnweisung)) {

			listeAnmelden(AnmeldeTyp.EMPFAENGER, anweisung.getEmpfaenger(), empfaengerAnmeldungen);
			listeAnmelden(AnmeldeTyp.SENKE, anweisung.getSenken(), senkeAnmeldungen);
			listeAnmelden(AnmeldeTyp.QUELLE, anweisung.getQuellen(), quelleAnmeldungen);
			listeAnmelden(AnmeldeTyp.SENDER, anweisung.getSender(), senderAnmeldungen);

			if (anweisung.getInitialAnweisungen().interpret(kontext) instanceof RueckgabeAnweisung) {
				beenden();
			} else {
				setRunning(true);
				kontext.getInterpreter().addAktivitaet(this);
				getKontext().getInterpreter().getTimer().schedule(new TimeoutChecker(this), 1000L, 1000L);
			}
		}
	}

	/**
	 * fügt dem der Aktivität eine Anmeldung des übergebenen Typs hinzu.
	 *
	 * @param typ
	 *            der Typ der Anmeldung
	 * @param eintrag
	 *            der anzumeldende Eintrag
	 */
	public void addAnmeldung(final AnmeldeTyp typ, final AtgAnmeldungsEintrag eintrag) {
		empfaengerAnmeldungen.add(eintrag);
		eintrag.anmelden(typ, this);
	}

	/**
	 * die Funktion aktualisiert den Marker für die Timeout-Behandlung mit dem
	 * aktuellen Zeitstempel.<br>
	 * Für das übergebene Systemobjekt und die Datenbeschreibung werden alle
	 * passenden Empfängeranmeldungen gesucht und der Zeitstempel für die
	 * Empfangstimeout auf Basis des aktuellen Zeitstempels aktualisiert.
	 *
	 * @param object
	 *            das Systemobjekt für das Daten empfangen worden
	 * @param dataDescription
	 *            die Datenbeschreibung für die empfangenen Daten
	 */
	private void aktualisiereAnmeldung(final SystemObject object, final DataDescription dataDescription) {
		final long zeit = System.currentTimeMillis();
		synchronized (empfaengerAnmeldungen) {
			for (final AnmeldungsEintrag eintrag : empfaengerAnmeldungen) {
				if (eintrag instanceof AtgAnmeldungsEintrag) {
					final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
					if (atgAnmeldung.matches(object, dataDescription.getAttributeGroup(),
							dataDescription.getAspect())) {
						atgAnmeldung.setAktualisierung(zeit);
					}
				}
			}
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Beim Beenden der Aktivität werden alle aktuellen Anmeldungen beim
	 * Datenverteiler abgemeldet. Der Status der AKtivität wird auf
	 * <i>inaktiv</i> gesetzt.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.anweisung.UdaAktivitaet#beenden()
	 */
	@Override
	public void beenden() {
		for (final AnmeldungsEintrag eintrag : empfaengerAnmeldungen) {
			eintrag.abmelden(AnmeldeTyp.EMPFAENGER, this);
		}
		for (final AnmeldungsEintrag eintrag : senkeAnmeldungen) {
			eintrag.abmelden(AnmeldeTyp.SENKE, this);
		}
		for (final AnmeldungsEintrag eintrag : senderAnmeldungen) {
			eintrag.abmelden(AnmeldeTyp.SENDER, this);
		}
		for (final AnmeldungsEintrag eintrag : quelleAnmeldungen) {
			eintrag.abmelden(AnmeldeTyp.QUELLE, this);
		}

		setRunning(false);
		synchronized (getKontext().getInterpreter()) {
			getKontext().getInterpreter().notifyAll();
		}
	}

	/**
	 * die Funktion führt den Fehlerbehandlungblock der Anmeldeanweisung aus.
	 * <br>
	 * Der übergebene Fehler wird der Fehlervariable im lokalen Kontext
	 * übergeben und anschließend die Anweisungen des Fehlerbehandlungsblocks
	 * ausgeführt.<br>
	 * Ist keine Fehlerbehandlungblock und damit keine Fehlerbehandlungsvariable
	 * definiert, wird der Fehler als unbehandelt weitergereicht und führt in
	 * der Regel zum Abbruch der Anmeldeanweisung.
	 *
	 * @param e
	 *            der zu behandelnde Fehler
	 * @return wahr, wenn die Anmeldung beendet werden soll
	 */
	synchronized boolean behandleFehler(final UdaFehler e) {
		final AnweisungsBlock fehlerAnweisungen = anweisung.getFehlerAnweisungen();
		final String fehlerVariable = anweisung.getFehlerVariable();
		if (fehlerVariable == null) {
			throw e;
		}
		if (e.getZeile() <= 0) {
			e.setZeile(anweisung.getZeile());
		}
		if (e.getSkriptName().length() <= 0) {
			e.setSkript(anweisung.getSkript());
		}
		getKontext().set(fehlerVariable, e);
		return (fehlerAnweisungen.interpret(getKontext()) instanceof RueckgabeAnweisung);
	}

	/**
	 * {@inheritDoc}.<br>
	 * setzt den Sendeanforderungsstaus für eine Sender- oder Quellenanmeldung.
	 * <br>
	 * Die Funktion ermittelt alle passenden Sender- und Quellenanmeldungen und
	 * setzt den Status der Anforderung auf den übergebenen Wert.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription, byte)
	 */
	@Override
	public void dataRequest(final SystemObject obj, final DataDescription dataDescription, final byte status) {
		for (final AnmeldungsEintrag eintrag : senderAnmeldungen) {
			if (eintrag instanceof AtgAnmeldungsEintrag) {
				final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
				if (atgAnmeldung.matches(obj, dataDescription.getAttributeGroup(), dataDescription.getAspect())) {
					atgAnmeldung.setRequestState(status);
				}
			}
		}
		for (final AnmeldungsEintrag eintrag : quelleAnmeldungen) {
			if (eintrag instanceof AtgAnmeldungsEintrag) {
				final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
				if (atgAnmeldung.matches(obj, dataDescription.getAttributeGroup(), dataDescription.getAspect())) {
					atgAnmeldung.setRequestState(status);
				}
			}
		}
	}

	/**
	 * liefert die Datenverteilerverbindung, über die die Anmeldung
	 * kommuniziert.
	 *
	 * @return die Verbindung
	 */
	public ClientDavInterface getConnection() {
		return connection;
	}

	/**
	 * liefert die Liste der aktuellen Anmeldungen als Empfänger.
	 *
	 * @return die Anmeldungen
	 */
	public Collection<AnmeldungsEintrag> getEmpfaengerAnmeldungen() {
		return empfaengerAnmeldungen;
	}

	/**
	 * liefert die Liste der aktuellen Anmeldungen als Quelle.
	 *
	 * @return die Anmeldungen
	 */
	public Collection<AnmeldungsEintrag> getQuelleAnmeldungen() {
		return quelleAnmeldungen;
	}

	/**
	 * liefert die Liste der aktuellen Anmeldungen als Sender.
	 *
	 * @return die Anmeldungen
	 */
	public Collection<AnmeldungsEintrag> getSenderAnmeldungen() {
		return senderAnmeldungen;
	}

	/**
	 * liefert die Liste der aktuellen Anmeldungen als Senke.
	 *
	 * @return die Anmeldungen
	 */
	public Collection<AnmeldungsEintrag> getSenkeAnmeldungen() {
		return senkeAnmeldungen;
	}

	/**
	 * {@inheritDoc}.<br>
	 * die Sendeanforderungssteuerung wird für alle Sende- und
	 * Quellenanmeldungen unterstützt.<br>
	 * Die Funktion ermittelt prüft ob eine passende Sender- oder
	 * Quellenanmeldungen existiert und liefert den entsprechenden Wert.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription)
	 */
	@Override
	public boolean isRequestSupported(final SystemObject obj, final DataDescription dataDescription) {
		boolean result = false;
		for (final AnmeldungsEintrag eintrag : quelleAnmeldungen) {
			if (eintrag instanceof AtgAnmeldungsEintrag) {
				final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
				if (atgAnmeldung.matches(obj, dataDescription.getAttributeGroup(), dataDescription.getAspect())) {
					result = true;
				}
			}
		}
		if (!result) {
			for (final AnmeldungsEintrag eintrag : senderAnmeldungen) {
				if (eintrag instanceof AtgAnmeldungsEintrag) {
					final AtgAnmeldungsEintrag atgAnmeldung = (AtgAnmeldungsEintrag) eintrag;
					if (atgAnmeldung.matches(obj, dataDescription.getAttributeGroup(), dataDescription.getAspect())) {
						result = true;
					}
				}
			}
		}
		return result;
	}

	/**
	 * die Funktion führt die Anmeldung für alle Ausdrücke für
	 * Anmeldungseinträge aus und trägt diese in die übergebene Liste der
	 * aktuellen Anmeldungen ein.<br>
	 * Treten bei der Anmeldung selbst Fehler auf, werden die Anweisungen des
	 * Fehlerbehandlungsbereichs ausgeführt. Unbehandelte Fehler bei der
	 * Auswertung der Anmeldeausdrücke führen zum Abbruch der
	 * Anmeldungsoperationen und werden nach aussen weitergereicht.
	 *
	 * @param typ
	 *            der Typ der ausführenden Anmeldung
	 * @param anmeldungen
	 *            die Liste der Ausdrücke für die Definition der Anmeldungen
	 * @param liste
	 *            die Liste der aktuellen Anmeldungen
	 */
	private void listeAnmelden(final AnmeldeTyp typ, final Collection<AnmeldeAusdruck> anmeldungen,
			final Collection<AnmeldungsEintrag> liste) {
		for (final Ausdruck eintrag : anmeldungen) {
			try {
				for (final AnmeldungsEintrag anmeldung : (AnmeldungsEintrag[]) eintrag.interpret(getKontext())) {
					try {
						liste.add(anmeldung);
						anmeldung.anmelden(typ, this);
					} catch (final KonfigurationsFehler e) {
						behandleFehler(e);
					} catch (final OneSubscriptionPerSendData e) {
						throw new RuntimeException(e);
					}
				}
			} catch (final KonfigurationsFehler e) {
				behandleFehler(e);
			}
		}
	}

	/**
	 * entfernt Empfängeranmeldungen von der Anmeldungs-Aktivität.<br>
	 * Es werden alle passenden Empfänger- und Senkenanmeldung ermittelt, vom
	 * Datenverteiler abgemeldet und aus der Liste der aktiven Anmeldungen
	 * entfernt.
	 *
	 * @param obj
	 *            das Objekt, welches das Systemobjekt beschreibt
	 * @param atgObj
	 *            das Objekt, welches die Attibutgruppe beschreibt
	 * @param aspObj
	 *            das Objekt, welches den Aspekt beschreibt
	 */
	public void removeEmpfaenger(final Object obj, final Object atgObj, final Object aspObj) {
		final SystemObject sysObj = (SystemObject) Anmeldung.getDavObject(connection, obj, SystemObject.class);
		final AttributeGroup atg = (AttributeGroup) Anmeldung.getDavObject(connection, atgObj, AttributeGroup.class);
		final Aspect asp = (Aspect) Anmeldung.getDavObject(connection, aspObj, Aspect.class);

		final Collection<AnmeldungsEintrag> removeList = new ArrayList<>();
		for (final AnmeldungsEintrag eintrag : empfaengerAnmeldungen) {
			if (eintrag.matches(sysObj, atg, asp)) {
				eintrag.abmelden(AnmeldeTyp.EMPFAENGER, this);
				removeList.add(eintrag);
			}
		}
		for (final AnmeldungsEintrag eintrag : removeList) {
			empfaengerAnmeldungen.remove(eintrag);
		}

		removeList.clear();
		for (final AnmeldungsEintrag eintrag : senkeAnmeldungen) {
			if (eintrag.matches(sysObj, atg, asp)) {
				eintrag.abmelden(AnmeldeTyp.SENKE, this);
				removeList.add(eintrag);
			}
		}
		for (final AnmeldungsEintrag eintrag : removeList) {
			senkeAnmeldungen.remove(eintrag);
		}

		removeList.clear();
	}

	/**
	 * entfernt Senderanmeldungen von der Anmeldungs-Aktivität.<br>
	 * Es werden alle passenden Sender- und Quellenanmeldung ermittelt, vom
	 * Datenverteiler abgemeldet und aus der Liste der aktiven Anmeldungen
	 * entfernt.
	 *
	 * @param obj
	 *            das Objekt, welches das Systemobjekt beschreibt
	 * @param atgObj
	 *            das Objekt, welches die Attibutgruppe beschreibt
	 * @param aspObj
	 *            das Objekt, welches den Aspekt beschreibt
	 */
	public void removeSender(final Object obj, final Object atgObj, final Object aspObj) {
		final SystemObject sysObj = (SystemObject) Anmeldung.getDavObject(connection, obj, SystemObject.class);
		final AttributeGroup atg = (AttributeGroup) Anmeldung.getDavObject(connection, atgObj, AttributeGroup.class);
		final Aspect asp = (Aspect) Anmeldung.getDavObject(connection, aspObj, Aspect.class);

		final Collection<AnmeldungsEintrag> removeList = new ArrayList<>();
		for (final AnmeldungsEintrag eintrag : senderAnmeldungen) {
			if (eintrag.matches(sysObj, atg, asp)) {
				eintrag.abmelden(AnmeldeTyp.SENDER, this);
				removeList.add(eintrag);
			}
		}
		for (final AnmeldungsEintrag eintrag : removeList) {
			senderAnmeldungen.remove(eintrag);
		}

		removeList.clear();
		for (final AnmeldungsEintrag eintrag : quelleAnmeldungen) {
			if (eintrag.matches(sysObj, atg, asp)) {
				eintrag.abmelden(AnmeldeTyp.QUELLE, this);
				removeList.add(eintrag);
			}
		}
		for (final AnmeldungsEintrag eintrag : removeList) {
			quelleAnmeldungen.remove(eintrag);
		}

		removeList.clear();
	}

	/**
	 * {@inheritDoc}.<br>
	 * Es werden alle empfangenen Mengenänderungen einzeln ausgewertet. Für
	 * jeden geänderten Eintrag wird ein Mengenänderungsobjekt erzeugt. Das
	 * erzeugte Objekt wird im lokalen Kontext in der Mengenänderungsvariable
	 * abgelegt und die Anweisungen des Mengenänderungsbereiches ausgeführt.<br>
	 * Tritt bei der Ausführung der Anweisungen des Mengenänderungsbereiches ein
	 * unbehandelter Fehler auf, wird der Fehlerbehandlungsbereich mit dem
	 * erhaltenen Fehlerobjekt ausgeführt.
	 *
	 * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet,
	 *      de.bsvrz.dav.daf.main.config.SystemObject[],
	 *      de.bsvrz.dav.daf.main.config.SystemObject[])
	 */
	@Override
	public void update(final MutableSet set, final SystemObject[] addedObjects, final SystemObject[] removedObjects) {
		final String variable = anweisung.getMengenAenderungsVariable();
		if (variable != null) {
			final AnweisungsBlock anweisungen = anweisung.getMengenAenderungsAnweisungen();
			final MengenObjekt menge = (MengenObjekt) DavDaten.erzeugeObjekt(getConnection(), set);
			for (final SystemObject obj : addedObjects) {
				try {
					getKontext().set(variable,
							new MengenAenderung(menge, DavDaten.erzeugeObjekt(getConnection(), obj), true));
					anweisungen.interpret(getKontext());
				} catch (final UdaFehler e) {
					behandleFehler(e);
				}
			}
			for (final SystemObject obj : removedObjects) {
				try {
					getKontext().set(variable,
							new MengenAenderung(menge, DavDaten.erzeugeObjekt(getConnection(), obj), false));
					anweisungen.interpret(getKontext());
				} catch (final UdaFehler e) {
					behandleFehler(e);
				}
			}
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Es werden alle empfangenen Datensätze einzeln ausgewertet. Enthält ein
	 * Datensatz Daten, wird der Aktualisierungsvariable im lokalen Kontext der
	 * Anmeldeanweisung der empfangene Datensatz zugewiesen und die Anweisungen
	 * des Aktualisierungsblockes ausgeführt. Zusätzlich wird der Marker für die
	 * Timeoutbehandlung der betroffenen Anmeldung zurückgesetzt.<br>
	 * Enthält der Datensatz keine Daten wird ein entprechendes Fehlerobjekt
	 * (Keine Quelle, Keine Daten, Keine Rechte) erzeugt, das Fehlerobjekt der
	 * Fehlerbehandlungsvariable zugewiesen und die Anweisungen des
	 * Fehlerbehandlungsblockes ausgeführt.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public void update(final ResultData[] results) {
		final String variable = anweisung.getAktualisierungsVariable();
		if (variable != null) {
			for (final ResultData result : results) {
				aktualisiereAnmeldung(result.getObject(), result.getDataDescription());
				if (result.isNoSourceAvailable()) {
					behandleFehler(new DatenZugriffsFehler(UdaFehlerSubtyp.QUELLE, "Keine Quelle angemeldet"));
				} else if (result.isNoDataAvailable()) {
					behandleFehler(new DatenZugriffsFehler(UdaFehlerSubtyp.DATEN, "Keine Daten verfügbar"));
				} else if (result.isNoRightsAvailable()) {
					behandleFehler(new DatenZugriffsFehler(UdaFehlerSubtyp.RECHTE, "Keine Rechte"));
				} else {
					getKontext().set(variable, DavDaten.erzeugeObjekt(getConnection(), result));
					final AnweisungsBlock block = anweisung.getAktualisierungsAnweisungen();
					if (block != null) {
						try {
							block.interpret(getKontext());
						} catch (final UdaFehler e) {
							behandleFehler(e);
						}
					}
				}
			}
		}
	}
}
