/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.anweisung.anmeldung;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.bsvrz.ibv.uda.interpreter.AbstractUdaHandler;
import de.bsvrz.ibv.uda.interpreter.daten.container.Feld;
import de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer;
import de.bsvrz.ibv.uda.interpreter.daten.container.UdaIterator;
import de.bsvrz.ibv.uda.interpreter.daten.dav.MengenObjekt;
import de.bsvrz.ibv.uda.interpreter.daten.zeit.ZeitDauer;
import de.bsvrz.sys.funclib.bitctrl.interpreter.HandlerValidation;
import de.bsvrz.sys.funclib.bitctrl.interpreter.NichtWert;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischerWert;

/**
 * der Handler zur Ausführung von Operationen mit einer Anmeldungs-Aktivität.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class AnmeldungHandler extends AbstractUdaHandler {

	/**
	 * {@inheritDoc}.
	 *
	 * @return liefert die Liste der unterstützten Operatoren
	 * @see de.bsvrz.sys.funclib.bitctrl.interpreter.Handler#getHandledOperators()
	 */
	@Override
	public Operator[] getHandledOperators() {
		return AnmeldungOperatoren.OPERATOREN;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion liefert unter folgenden Bedingungen den Wert <i>true</i>:
	 * <ul>
	 * <li><b>senkeAnmeldungen:</b> die Operation hat einen Operanden und dieser
	 * ist eine Anmeldungsaktivität</li>
	 * <li><b>empfängerAnmeldungen:</b> die Operation hat einen Operanden und
	 * dieser ist eine Anmeldungsaktivität</li>
	 * <li><b>senderAnmeldungen:</b> die Operation hat einen Operanden und
	 * dieser ist eine Anmeldungsaktivität</li>
	 * <li><b>quelleAnmeldungen:</b> die Operation hat einen Operanden und
	 * dieser ist eine Anmeldungsaktivität</li>
	 * <li><b>lesen:</b> die Operation hat 5 oder 6 Operanden. Der erste Operand
	 * ist eine Anmeldungsaktivität, der zweite ist ein logischer Wert, der
	 * optionale sechste ist eine Zeitdauer. Die übrigen Argumente werden erst
	 * zur Laufzeit. d.h. beim Ausführen der Operation geprüft.</li>
	 * <li><b>nichtLesen:</b> die Operation hat 4 Operanden und der erste
	 * Operand ist eine Anmeldungsaktivität. Die übrigen Argumente werden erst
	 * zur Laufzeit. d.h. beim Ausführen der Operation geprüft.</li>
	 * <li><b>schreiben:</b> die Operation hat 5 Operanden. Der erste Operand
	 * ist eine Anmeldungsaktivität, der zweite ist ein logischer Wert. Die
	 * übrigen Argumente werden erst zur Laufzeit. d.h. beim Ausführen der
	 * Operation geprüft.</li>
	 * <li><b>nichtSchreiben:</b> die Operation hat 4 Operanden und der erste
	 * Operand ist eine Anmeldungsaktivität. Die übrigen Argumente werden erst
	 * zur Laufzeit. d.h. beim Ausführen der Operation geprüft.</li>
	 * </ul>
	 *
	 * @see de.bsvrz.sys.funclib.bitctrl.interpreter.Handler#validiereHandler(de.bsvrz.sys.funclib.bitctrl.interpreter.Operator,
	 *      java.util.List)
	 */
	@Override
	public HandlerValidation validiereHandler(final Operator operator, final List<?> operanden) {
		HandlerValidation result = null;
		if (AnmeldungOperatoren.SENKE_ANMELDUNGEN.equals(operator)) {
			result = pruefeAnmeldungsAbfrage(operanden);
		} else if (AnmeldungOperatoren.EMPFAENGER_ANMELDUNGEN.equals(operator)) {
			result = pruefeAnmeldungsAbfrage(operanden);
		} else if (AnmeldungOperatoren.SENDER_ANMELDUNGEN.equals(operator)) {
			result = pruefeAnmeldungsAbfrage(operanden);
		} else if (AnmeldungOperatoren.QUELLE_ANMELDUNGEN.equals(operator)) {
			result = pruefeAnmeldungsAbfrage(operanden);
		} else if (AnmeldungOperatoren.LESEN.equals(operator)) {
			result = pruefeLeseAnmeldung(operanden);
		} else if (AnmeldungOperatoren.SCHREIBEN.equals(operator)) {
			result = pruefeSchreibAnmeldung(operanden);
		} else if (AnmeldungOperatoren.NICHT_LESEN.equals(operator)) {
			result = pruefeAbmeldung(operanden);
		} else if (AnmeldungOperatoren.NICHT_SCHREIBEN.equals(operator)) {
			result = pruefeAbmeldung(operanden);
		}

		return result;
	}

	/**
	 * die Funktion ermittelt, ob die übergebenen Argumente für eine
	 * Schreibanmeldung, d.h. eine Anmeldung als Sender oder Quelle, gültig
	 * sind.<br>
	 * Für eine gültige Schreibanmeldung werden 5 Argumente erwartet.
	 * <ul>
	 * <li>die Anmeldeaktivität, der eine Schreibanmeldung hinzugefügt werden
	 * soll.</li>
	 * <li>ein logischer Wert, der festlegt, ob die Anmeldung als Quelle oder
	 * Sender erfolgen soll,</li>
	 * <li>ein Objekt, das ein Systemobjekt repräsentiert</li>
	 * <li>ein Objekt, das eine Attributgruppe repräsentiert</li>
	 * <li>ein Objekt, das einen Aspekt repräsentiert</li>
	 * </ul>
	 * Die Objekte zur Bildung der Datenbeschreibung werden nicht überprüft.
	 *
	 * @param operanden
	 *            die zu prüfenden Operanden
	 * @return das Ergebnis der Überprüfung
	 */
	private HandlerValidation pruefeSchreibAnmeldung(final List<?> operanden) {
		boolean anzahlOk = false;
		boolean typOk = false;
		if (operanden.size() == 5) {
			anzahlOk = true;
			if (operanden.get(0) instanceof Anmeldung) {
				if (operanden.get(1) instanceof LogischerWert) {
					typOk = true;
				}
			}
		}
		return new HandlerValidation(anzahlOk, typOk);
	}

	/**
	 * die Funktion ermittelt, ob die übergebenen Argumente für eine
	 * Leseanmeldung, d.h. eine Anmeldung als Empfänger oder Senke, gültig sind.
	 * <br>
	 * Für eine gültige Schreibanmeldung werden 5 oder 6 Argumente erwartet.
	 * <ul>
	 * <li>die Anmeldeaktivität, der eine Schreibanmeldung hinzugefügt werden
	 * soll.</li>
	 * <li>ein logischer Wert, der festlegt, ob die Anmeldung als Senke oder
	 * Empfänger erfolgen soll,</li>
	 * <li>ein Objekt, das ein Systemobjekt repräsentiert</li>
	 * <li>ein Objekt, das eine Attributgruppe repräsentiert</li>
	 * <li>ein Objekt, das einen Aspekt repräsentiert</li>
	 * <li>eine optionale Zeitdauer für den Timeoutwert der Anmeldung</li>
	 * </ul>
	 * Die Objekte zur Bildung der Datenbeschreibung werden nicht überprüft.
	 *
	 * @param operanden
	 *            die zu prüfenden Operanden
	 * @return das Ergebnis der Überprüfung
	 */
	private HandlerValidation pruefeLeseAnmeldung(final List<?> operanden) {
		boolean anzahlOk = false;
		boolean typOk = false;
		if ((operanden.size() == 5) || (operanden.size() == 6)) {
			anzahlOk = true;
			if (operanden.get(0) instanceof Anmeldung) {
				if (operanden.get(1) instanceof LogischerWert) {
					if (operanden.size() == 6) {
						if (operanden.get(5) instanceof ZeitDauer) {
							typOk = true;
						}
					} else {
						typOk = true;
					}
				}
			}
		}

		return new HandlerValidation(anzahlOk, typOk);
	}

	/**
	 * die Funktion ermittelt, ob die übergebenen Argumente für eine Abmeldung
	 * gültig sind.<br>
	 * Für eine gültige Abmeldung werden 4 Argumente erwartet.
	 * <ul>
	 * <li>die Anmeldeaktivität, der eine Schreibanmeldung hinzugefügt werden
	 * soll.</li>
	 * <li>ein Objekt, das ein Systemobjekt repräsentiert</li>
	 * <li>ein Objekt, das eine Attributgruppe repräsentiert</li>
	 * <li>ein Objekt, das einen Aspekt repräsentiert</li>
	 * </ul>
	 * Die Objekte zur Bildung der Datenbeschreibung werden nicht überprüft.
	 *
	 * @param operanden
	 *            die zu prüfenden Operanden
	 * @return das Ergebnis der Überprüfung
	 */
	private HandlerValidation pruefeAbmeldung(final List<?> operanden) {
		boolean anzahlOk = false;
		boolean typOk = false;

		if (operanden.size() == 4) {
			anzahlOk = true;
			if (operanden.get(0) instanceof Anmeldung) {
				typOk = true;
			}
		}
		return new HandlerValidation(anzahlOk, typOk);
	}

	/**
	 * die Funktion ermittelt, ob die übergebenen Argumente für die Abfrage
	 * einer Menge von Anmeldungen gültig sind.<br>
	 * Für eine gültige Abfrage wird ein Argument erwartet.
	 * <ul>
	 * <li>die Anmeldeaktivität, der eine Schreibanmeldung hinzugefügt werden
	 * soll.</li>
	 * </ul>
	 *
	 * @param operanden
	 *            die zu prüfenden Operanden
	 * @return das Ergebnis der Überprüfung
	 */
	private HandlerValidation pruefeAnmeldungsAbfrage(final List<?> operanden) {
		boolean anzahlOk = false;
		boolean typOk = false;
		if (operanden.size() == 1) {
			anzahlOk = true;
			if (operanden.get(0) instanceof Anmeldung) {
				typOk = true;
			}
		}
		return new HandlerValidation(anzahlOk, typOk);
	}

	/**
	 * {@inheritDoc}.<br>
	 * Folgende Operatoren können ausgeführt werden:
	 * <ul>
	 * <li>senkeAnmeldungen: ein Uda-Feld mit den ermittelten Senke-Anmeldungen
	 * </li>
	 * <li>empfängerAnmeldungen: ein Uda-Feld mit den ermittelten
	 * Empfänger-Anmeldungen</li>
	 * <li>senderAnmeldungen: ein Uda-Feld mit den ermittelten
	 * Sender-Anmeldungen</li>
	 * <li>quelleAnmeldungen: ein Uda-Feld mit den ermittelten
	 * Quelle-Anmeldungen</li>
	 *
	 * <li>lesen: den hinzugefügten Anmeldungseintrag</li>
	 * <li>nichtLesen: den Wert "undefiniert"</li>
	 * <li>schreiben: den hinzugefügten Anmeldungseintrag</li>
	 * <li>nichtSchreiben: den Wert "undefiniert"</li>
	 * </ul>
	 *
	 * @see de.bsvrz.sys.funclib.bitctrl.interpreter.Handler#perform(de.bsvrz.sys.funclib.bitctrl.interpreter.Operator,
	 *      java.util.List)
	 */
	@Override
	public Object perform(final Operator operator, final List<?> operanden) {

		Object result = null;

		erzeugeHandlerFehler(operator, operanden);

		if (AnmeldungOperatoren.SENKE_ANMELDUNGEN.equals(operator)) {
			result = new Feld(((Anmeldung) operanden.get(0)).getSenkeAnmeldungen());
		} else if (AnmeldungOperatoren.EMPFAENGER_ANMELDUNGEN.equals(operator)) {
			result = new Feld(((Anmeldung) operanden.get(0)).getEmpfaengerAnmeldungen());
		} else if (AnmeldungOperatoren.SENDER_ANMELDUNGEN.equals(operator)) {
			result = new Feld(((Anmeldung) operanden.get(0)).getSenderAnmeldungen());
		} else if (AnmeldungOperatoren.QUELLE_ANMELDUNGEN.equals(operator)) {
			result = new Feld(((Anmeldung) operanden.get(0)).getQuelleAnmeldungen());
		} else if (AnmeldungOperatoren.LESEN.equals(operator)) {
			zumLesenAnmelden(operanden);
		} else if (AnmeldungOperatoren.SCHREIBEN.equals(operator)) {
			zumSchreibenAnmelden(operanden);
		} else if (AnmeldungOperatoren.NICHT_LESEN.equals(operator)) {
			final Anmeldung anmeldung = (Anmeldung) operanden.get(0);
			for (final Object objekt : getObjectListe(operanden.get(2))) {
				anmeldung.removeEmpfaenger(operanden.get(1), objekt, operanden.get(3));
			}
		} else if (AnmeldungOperatoren.NICHT_SCHREIBEN.equals(operator)) {
			final Anmeldung anmeldung = (Anmeldung) operanden.get(0);
			for (final Object objekt : getObjectListe(operanden.get(2))) {
				anmeldung.removeSender(operanden.get(1), objekt, operanden.get(3));
			}
		}

		if (result == null) {
			result = NichtWert.NICHTWERT;
		}

		return result;
	}

	/**
	 * die Funktion erzeugt eine neue Anmeldung als Empfänger oder Senke und
	 * registriert diese bei der übergebenen Anmeldungs-Aktivität.
	 *
	 * @param operanden
	 *            die Operanden
	 */
	private void zumLesenAnmelden(final List<?> operanden) {
		final Anmeldung anmeldung = (Anmeldung) operanden.get(0);
		final boolean senke = ((LogischerWert) operanden.get(1)).get();
		ZeitDauer dauer = null;
		if (operanden.size() == 6) {
			dauer = (ZeitDauer) operanden.get(5);
		}

		final Collection<Object> objectListe = getObjectListe(operanden.get(2));
		final Object atg = operanden.get(3);
		final Object asp = operanden.get(4);

		for (final Object obj : objectListe) {
			final AtgAnmeldungsEintrag eintrag = new AtgAnmeldungsEintrag(anmeldung.getConnection(), obj, atg, asp,
					dauer);
			if (senke) {
				anmeldung.addAnmeldung(AnmeldeTyp.SENKE, eintrag);
			} else {
				anmeldung.addAnmeldung(AnmeldeTyp.EMPFAENGER, eintrag);
			}
		}
	}

	/**
	 * ermittelt die Liste der Objekte, die sich aus einem Anmeldungsobjekt
	 * ergibt.
	 *
	 * @param objekt
	 *            das Objekt
	 * @return die Liste
	 */
	private Collection<Object> getObjectListe(final Object objekt) {
		final Collection<Object> result = new ArrayList<>();
		if (objekt instanceof UdaContainer) {
			final UdaIterator iterator = ((UdaContainer) objekt).getUdaIterator(UdaIterator.ANFANG);
			while (iterator.weiter().get()) {
				result.add(iterator.getElement());
			}
		} else if (objekt instanceof MengenObjekt) {
			final UdaIterator iterator = ((MengenObjekt) objekt).getMengenObjekte().getUdaIterator(UdaIterator.ANFANG);
			while (iterator.weiter().get()) {
				result.add(iterator.getElement());
			}
		} else {
			result.add(objekt);
		}

		return result;
	}

	/**
	 * die Funktion erzeugt eine neue Anmeldung als Sender oder Quelle und
	 * registriert diese bei der übergebenen Anmeldungs-Aktivität.
	 *
	 * @param operanden
	 *            die Operanden
	 */
	private void zumSchreibenAnmelden(final List<?> operanden) {

		final Anmeldung anmeldung = (Anmeldung) operanden.get(0);
		final boolean quelle = ((LogischerWert) operanden.get(1)).get();

		final Collection<Object> objectListe = getObjectListe(operanden.get(2));
		final Object atg = operanden.get(3);
		final Object asp = operanden.get(4);

		for (final Object obj : objectListe) {
			final AtgAnmeldungsEintrag eintrag = new AtgAnmeldungsEintrag(anmeldung.getConnection(), obj, atg, asp,
					null);
			if (quelle) {
				anmeldung.addAnmeldung(AnmeldeTyp.QUELLE, eintrag);
			} else {
				anmeldung.addAnmeldung(AnmeldeTyp.SENDER, eintrag);
			}
		}
	}
}
