/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.daten.container;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.bsvrz.ibv.uda.interpreter.Kopierbar;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ContainerZugriffsFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.sys.funclib.bitctrl.interpreter.NichtWert;

/**
 * Repräsentation eines Felds als spezielle Ausprägung eines
 * {@link UdaContainer}.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class Feld extends UdaContainer implements Kopierbar {
	/**
	 * die Daten des Feldes.
	 */
	private final List<Object> feld = new ArrayList<>();

	/**
	 * Standard-Konstruktor.<br>
	 * Der Konstruktor erzeugt ein leeres Feld.
	 */
	public Feld() {
		super();
	}

	/**
	 * Konstruktor mit Angabe einer Java-Collection.<br>
	 * Der Konstruktor erzeugt ein Feld und kopiert alle Referenzen aus der
	 * übergebenen Collection. Es werden lediglich die Referenzen des Feldes
	 * kopiert, d.h. es werden keine Kopien der innerhalb des Feldes
	 * gespeicherten Elemente erzeugt.
	 *
	 * @param quelle
	 *            die Collection mit den initialien Objektreferenzen des Felds
	 */
	public Feld(final Collection<? extends Object> quelle) {
		super();
		feld.addAll(quelle);
	}

	/**
	 * Konstruktor mit Angabe der Grösse.<br>
	 * Der Konstruktor erzeugt ein Feld mit der angegebenen Größe. Alle ELemente
	 * werden mit dem Nichtwert "undefiniert" initialisiert.
	 *
	 * @param groesse
	 *            die initiale Grösse des Feldes
	 */
	public Feld(final Number groesse) {
		super();
		for (int idx = 0; idx < groesse.intValue(); idx++) {
			anhaengen(-1, NichtWert.NICHTWERT);
		}
		if (groesse.intValue() < 0) {
			throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE, "Initialgröße des Feldes muss 0 oder positiv sein");
		}
	}

	/**
	 * Konstruktor mit Angabe eines Quellcontainers.<br>
	 * Der Konstruktor erzeugt ein Feld und kopiert alle Referenzen aus dem
	 * übergebenen Feld. Es werden lediglich die Referenzen des Feldes kopiert,
	 * d.h. es werden keine Kopien der innerhalb des Feldes gespeicherten
	 * Elemente erzeugt.
	 *
	 * @param container
	 *            der Container mit den initialien Objektreferenzen des Felds
	 */
	public Feld(final UdaContainer container) {
		super();
		feld.addAll(container.getElemente());
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#elementEinfuegen(int,
	 *      java.lang.Object)
	 */
	@Override
	void elementEinfuegen(final int index, final Object daten) {
		if (index < 0) {
			throw new ContainerZugriffsFehler(UdaFehlerSubtyp.INDEX);
		}
		if (index < getGroesse()) {
			feld.add(index, daten);
		} else {
			feld.add(daten);
		}
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#elementEinfuegen(java.lang.String,
	 *      java.lang.Object)
	 */
	@Override
	void elementEinfuegen(final String schluessel, final Object daten) {
		throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE);
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#elementLoeschen(int)
	 */
	@Override
	Object elementLoeschen(final int index) {
		final Object element = getContainerElement(index);
		feld.remove(index - 1);
		return element;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#elementLoeschen(java.lang.String)
	 */
	@Override
	Object elementLoeschen(final String schluessel) {
		return new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE);
	}

	/**
	 * {@inheritDoc}.<br>
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.Kopierbar#erzeugeKopie()
	 */
	@Override
	public Kopierbar erzeugeKopie() {
		final Feld result = new Feld();
		result.feld.addAll(feld);
		return result;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.Container#getContainerElement(int)
	 * @see #setContainerElement(int, Object)
	 */
	@Override
	public Object getContainerElement(final int index) {

		Object result = null;

		int zugriffsIndex = index;
		if (zugriffsIndex < 0) {
			zugriffsIndex = getGroesse() + 1 + zugriffsIndex;
		}
		if ((zugriffsIndex <= 0) || (zugriffsIndex > getGroesse())) {
			result = new ContainerZugriffsFehler(UdaFehlerSubtyp.INDEX);
		} else {
			result = feld.get(zugriffsIndex - 1);
		}
		return result;
	}

	/**
	 * liefert die Elemente, die in dem Feld gespeichert sind.<br>
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#getElemente()
	 */
	@Override
	public List<Object> getElemente() {
		return feld;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#getGroesse()
	 */
	@Override
	public int getGroesse() {
		return feld.size();
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#getSchluessel(int)
	 */
	@Override
	String getSchluessel(final int position) {
		return Integer.toString(position);
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.UdaContainer#getSchluesselPosition(java.lang.Object)
	 */
	@Override
	int getSchluesselPosition(final Object schluessel) {
		if (schluessel instanceof Number) {
			return ((Number) schluessel).intValue();
		}
		throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE);
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.daten.container.Container#setContainerElement(int,
	 *      java.lang.Object)
	 * @see #getContainerElement(int)
	 */
	@Override
	public void setContainerElement(final int index, final Object daten) {
		int zugriffsIndex = index;
		if (zugriffsIndex < 0) {
			zugriffsIndex = getGroesse() + 1 + zugriffsIndex;
		}
		if (zugriffsIndex <= 0) {
			throw new ContainerZugriffsFehler(UdaFehlerSubtyp.INDEX);
		}

		while (feld.size() < zugriffsIndex) {
			feld.add(NichtWert.NICHTWERT);
		}
		feld.set(zugriffsIndex - 1, daten);
	}

}
