/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.verwaltung;

import java.io.Externalizable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.uda.interpreter.InterpreterListener;
import de.bsvrz.ibv.uda.interpreter.UdaInterpreter;
import de.bsvrz.ibv.uda.interpreter.UdaInterpreterException;
import de.bsvrz.ibv.uda.interpreter.UdaParserFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehler;
import de.bsvrz.ibv.uda.uda.UdaMessage;
import de.bsvrz.ibv.uda.uda.data.Historie;
import de.bsvrz.ibv.uda.uda.data.LaufKommandoAntwort;
import de.bsvrz.ibv.uda.uda.data.Skript;
import de.bsvrz.ibv.uda.uda.data.SkriptLauf;
import de.bsvrz.ibv.uda.uda.data.SkriptLaufKommando;
import de.bsvrz.ibv.uda.uda.data.SkriptLaufStatus;
import de.bsvrz.ibv.uda.uda.data.SkriptZustand;
import de.bsvrz.ibv.uda.uda.data.StartTyp;
import de.bsvrz.ibv.uda.uda.data.StartZyklus;
import de.bsvrz.ibv.uda.uda.data.UDA;
import de.bsvrz.ibv.uda.verwaltung.protokoll.DateiProtokoll;
import de.bsvrz.ibv.uda.verwaltung.protokoll.DavProtokoll;
import de.bsvrz.ibv.uda.verwaltung.protokoll.KonsoleProtokoll;
import de.bsvrz.ibv.uda.verwaltung.protokoll.LogProtokoll;
import de.bsvrz.ibv.uda.verwaltung.protokoll.Protokoll;
import de.bsvrz.ibv.uda.verwaltung.protokoll.ProtokollEintrag;
import de.bsvrz.ibv.uda.verwaltung.protokoll.ProtokollTyp;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

/**
 * Repräsentation eines Skriptlaufobjekts innerhalb der Verwaltung der
 * "Umfassenden Datenanalyse".
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class ServerSkriptLauf extends SkriptLauf
implements ClientReceiverInterface, ClientSenderInterface, Externalizable, InterpreterListener {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Felder zur Serialisierung der Elemente des Objekts.<br>
	 *
	 * @author peuker
	 */
	public enum SkriptLaufFeld {
		/**
		 * Datenfeld für den Name des Skriptlaufs.
		 */
		NAME,
		/**
		 * Datenfeld für die Beschreibung des Skriptlaufs.
		 */
		BESCHREIBUNG,
		/**
		 * Datenfeld für den Namen des Benutzers unter dem der Skriptlauf
		 * auszuführen ist.
		 */
		BENUTZER,
		/**
		 * Datenfeld für das Passwort für die Anmeldung bei der Ausführung des
		 * Skripts.
		 */
		PASSWD,
		/**
		 * Datenfeld für das zugeordnete Skript.
		 */
		SKRIPT,
		/**
		 * Datenfeld für die Datenverteilerparameter.
		 */
		DAVPARAMETER,
		/**
		 * Datenfeld für die Skript-Aufrufparameter.
		 */
		SKRIPTPARAMETER,
		/**
		 * Datenfeld für den Starttyp des Skriptlaufs.
		 */
		STARTTYP,
		/**
		 * Datenfeld für den Zyklus für die Ausführung des Skriptlaufs.
		 */
		ZYKLUS,
		/**
		 * Datenfeld für die Speicherung der Historie für die Bearbeitung des
		 * Skriptlaufs.
		 */
		HISTORIE, /**
		 * Datenfeld für die Speicherung der Historie für die
		 * Bearbeitung des Skriptlaufs.
		 */
		PROTOKOLL
	}

	/**
	 * Versions-ID für die Serialisierung.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * die Datei, in der die Daten des Skriptlaufobjekts im Dateisystems des
	 * Servermoduls der "Umfassenden Datenanalyse" abgelegt werden.
	 */
	private File laufDatei;

	/**
	 * das Modul, in dem der Skriptlauf verwaltet wird.
	 */
	private ServerUdaModul modul;

	/**
	 * das Systemobjekt, das den Skriptlauf innerhalb der Konfiguration des
	 * Datenverteilers repräsentiert.
	 */
	private SystemObject objekt;

	/**
	 * die Datenbeschreibung für die Veröffentlichung der Daten des
	 * Skriptlaufobjekts.
	 */
	private DataDescription istDesc;

	/**
	 * die Datenbeschreibung zum Empfang von Daten zur Änderung des
	 * Skriptlaufobjekts.
	 */
	private DataDescription vorgabeDesc;

	/**
	 * die Datenbeschreibung zum Empfang von Anforderungen zur Ausführung des
	 * Skriptlaufs.
	 */
	private DataDescription kommandoDesc;

	/**
	 * die Datenbeschreibung zum Versand vin Antworten auf Anforderungen zur
	 * Ausführung des Skriptlaufs.
	 */
	private DataDescription kommandoRespDesc;

	/**
	 * die Datenbeschreibung zur Veröffentlichung des aktuellen
	 * Skriptlaufzustandes.
	 */
	private DataDescription statusDesc;

	/**
	 * die Datenbeschreibung zum Versenden von DavProtokoll-Daten.
	 */
	private DataDescription loggerDesc;

	/**
	 * die Datenbeschreibung zum Versenden des Skriptlaufergebnisses.
	 */
	private DataDescription resultDesc;

	/**
	 * der Interpreter, der den Skriptlauf ausführt.
	 */
	private UdaInterpreter laufInterpreter;

	/**
	 * der letzte Startzeitpunkt des Skripts.
	 */
	private long letzterStartZeitPunkt;

	/**
	 * Konstruktor.<br>
	 * Die Funtion erzeugt ein Skriptlaufobjekt für das übergebene UdaModul und
	 * das spezifizierte Systemobjekt.
	 *
	 * @param modul
	 *            das UdaModul, zu dem das Skript gehört
	 * @param obj
	 *            das Systemobjekt
	 * @throws UdaServerException
	 *             das Objekt konnte nicht angelegt werden
	 */
	public ServerSkriptLauf(final ServerUdaModul modul, final SystemObject obj) throws UdaServerException {

		try {
			this.modul = modul;
			this.objekt = obj;

			laufDatei = new File(modul.getModulVerzeichnis(), "lauf");
			if (!laufDatei.exists()) {
				laufDatei.mkdirs();
			}
			laufDatei = new File(laufDatei, obj.getPid());

			initialisiereVerbindung();
			ladeDaten(laufDatei);
			publiziereDaten();
			publiziereStatus();
		} catch (final IOException e) {
			throw new UdaServerException("Das Skriptlaufobjekt konnte nicht erstellt werden", e);
		}
	}

	/**
	 * die Funktion übernimmt die empfangenen Daten in die internen
	 * Datenstrukturen des Objekts und sichert diese in Datesystem des Servers
	 * der "Umfassenden Datenanalyse".
	 *
	 * @param data
	 *            die neuen Daten
	 * @throws UdaServerException
	 *             die daten konnten nicht korrekt eingelesen werden.
	 */
	private void aktualisiereDaten(final Data data) throws UdaServerException {
		setName(data.getTextValue(UDA.LAUF_ATT_NAME).getText());
		setBeschreibung(data.getTextValue(UDA.LAUF_ATT_BESCHREIBUNG).getText());
		setSkript(modul.getSkript(data.getReferenceValue(UDA.LAUF_ATT_SKRIPT).getSystemObject()));
		clearParameter();

		final Data.Array parameter = data.getArray(UDA.LAUF_ATT_PARAMETER);
		clearParameter();
		for (int idx = 0; idx < parameter.getLength(); idx++) {
			final Data eintrag = parameter.getItem(idx);
			final String name = eintrag.getTextValue(UDA.ATL_PARAMETER_ATT_NAME).getText();
			final String wert = eintrag.getTextValue(UDA.ATL_PARAMETER_ATT_WERT).getText();
			setParameter(name, wert);
		}

		setBenutzer(data.getTextValue(UDA.LAUF_ATT_BENUTZER).getText());

		final Data.NumberArray passwdArray = data.getUnscaledArray(UDA.LAUF_ATT_PASSWORT);
		if (passwdArray.getLength() > 0) {
			final byte[] passwdData = new byte[passwdArray.getLength()];
			for (int idx = 0; idx < passwdArray.getLength(); idx++) {
				passwdData[idx] = passwdArray.getValue(idx).byteValue();
			}
			setPasswd(getModul().decrypt(passwdData));
		}

		setStartTyp(StartTyp.getZustand(data.getUnscaledValue(UDA.LAUF_ATT_STARTTYP).intValue()));

		getZyklus().setStartZeitPunkt(data.getItem(UDA.LAUF_ATT_STARTZYKLUS)
				.getTimeValue(UDA.ATL_STARTZYKLUS_ATT_STARTZEITPUNKT).getMillis());
		getZyklus().setIntervall(
				data.getItem(UDA.LAUF_ATT_STARTZYKLUS).getUnscaledValue(UDA.ATL_STARTZYKLUS_ATT_INTERVALL).longValue());

		final Data.Array protokollArray = data.getArray(UDA.LAUF_ATT_PROTOKOLL);
		clearProtokolle();
		for (int idx = 0; idx < protokollArray.getLength(); idx++) {
			final Data eintrag = protokollArray.getItem(idx);
			Protokoll protokoll = null;
			final int typ = eintrag.getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_TYP).intValue();
			if (typ == ProtokollTyp.DATEI.getCode()) {
				protokoll = new DateiProtokoll();
			} else if (typ == ProtokollTyp.DAV.getCode()) {
				protokoll = new DavProtokoll();
			} else if (typ == ProtokollTyp.KONSOLE.getCode()) {
				protokoll = new KonsoleProtokoll();
			} else if (typ == ProtokollTyp.LOGGING.getCode()) {
				protokoll = new LogProtokoll();
			}

			if (protokoll != null) {
				protokoll.setName(eintrag.getTextValue(UDA.ATL_PROTOKOLL_ATT_NAME).getText());
				protokoll.setAnzahl(eintrag.getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_ANZAHL).longValue());
				protokoll.setMaxSize(eintrag.getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_MAXGROESSE).longValue());
				addProtokoll(protokoll);
			}
		}

		final Data historieData = data.getArray(UDA.LAUF_ATT_HISTORIE).getItem(0);
		if (historieData != null) {
			final long zeitPunkt = historieData.getTimeValue(UDA.ATL_HISTORIE_ATT_AENDERUNGSDATUM).getMillis();
			final String name = historieData.getTextValue(UDA.ATL_HISTORIE_ATT_NAME).getText();
			final String grund = historieData.getTextValue(UDA.ATL_HISTORIE_ATT_GRUND).getText();

			addHistorie(new Historie(zeitPunkt, name, grund));
		}

	}

	/**
	 * hält die Ausführung des Skriptlaufs an.
	 *
	 * @return das Ergebnis des Kommandoausführung
	 */
	public final LaufKommandoAntwort anhalten() {
		LaufKommandoAntwort antwort = null;
		final ServerSkript skript = getSkript();
		if (skript == null) {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
					"Dem Skriptlauf wurde kein Skript zugeordnet.");
		} else if (SkriptLaufStatus.AKTIV.equals(getStatus())) {
			if (getInterpreter() != null) {
				getInterpreter().anhalten();
			}
			setStatus(SkriptLaufStatus.ANGEHALTEN);
			publiziereStatus();
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.OK, "Skript wird angehalten.");
		} else {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.NICHT_AKTIV, "Skript ist nicht aktiv.");
		}

		return antwort;
	}

	/**
	 * veröffentlicht die übergebenen Ausgabedaten über die entsprechende
	 * Attributgruppe des Skriptlaufs.
	 *
	 * @param meldungsListe
	 *            die Liste der auszugebenden Texte
	 */
	public void ausgeben(final List<String> meldungsListe) {
		final Data daten = getModul().getVerbindung().createData(loggerDesc.getAttributeGroup());

		final Data.Array array = daten.getArray("Meldung");
		array.setLength(meldungsListe.size());
		for (int idx = 0; idx < meldungsListe.size(); idx++) {
			String text = meldungsListe.get(idx);
			if (text.length() > 1000) {
				text = text.substring(0, 1000) + " ...";
			}
			array.getTextValue(idx).setText(text);
		}
		try {
			getModul().getVerbindung().sendData(new ResultData(objekt, loggerDesc, System.currentTimeMillis(), daten));
		} catch (final DataNotSubscribedException e) {
			// sollte nicht auftreten bei Anmeldung als Quelle
			e.printStackTrace();
		} catch (final SendSubscriptionNotConfirmed e) {
			// sollte nicht auftreten bei Anmeldung als Quelle
			e.printStackTrace();
		}
	}

	/**
	 * beendet die Ausführung des Skriptlaufs.
	 *
	 * @return das Ergebnis des Kommandoausführung
	 */
	public final LaufKommandoAntwort beenden() {
		LaufKommandoAntwort antwort = null;
		final ServerSkript skript = getSkript();
		if (skript == null) {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
					"Dem Skriptlauf wurde kein Skript zugeordnet.");
		} else if (SkriptLaufStatus.AKTIV.equals(getStatus()) || SkriptLaufStatus.ANGEHALTEN.equals(getStatus())) {
			if (getInterpreter() != null) {
				getInterpreter().beenden();
			}
			setStatus(SkriptLaufStatus.PASSIV);
			publiziereStatus();
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.OK, "Skript wird beendet.");
		} else {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.NICHT_AKTIV, "Skript ist nicht aktiv.");
		}

		return antwort;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.interpreter.InterpreterListener#beendet(de.bsvrz.ibv.uda.interpreter.UdaInterpreter)
	 */
	@Override
	public void beendet(final UdaInterpreter interpreter) {
		if (interpreter.equals(laufInterpreter)) {
			setStatus(SkriptLaufStatus.BEENDET);
			publiziereStatus();
		}
	}

	/**
	 * ermittelt, ob der Skriptlauf zum übergebenen Zeitpunkt ausgeführt werden
	 * soll.
	 *
	 * @param startZeitPunkt
	 *            der gewünschte Startzeitpunkt
	 * @return <i>true</i>, wenn die Ausführung erfolgen kann
	 */
	public boolean bereitZurAutomatischenAusfuehrung(final long startZeitPunkt) {
		boolean result = false;

		if ((getSkript() != null) && getSkript().isFreigegeben()) {
			if (StartTyp.SOFORT.equals(getStartTyp())) {
				if (SkriptLaufStatus.PASSIV.equals(getStatus())) {
					result = true;
				}
			} else if (StartTyp.ZYKLISCH.equals(getStartTyp())) {
				if (startZeitPunkt >= getZyklus().getNaechstenStartZeitPunkt(letzterStartZeitPunkt)) {
					if (SkriptLaufStatus.PASSIV.equals(getStatus()) || SkriptLaufStatus.BEENDET.equals(getStatus())) {
						result = true;
					}
				}
			}
		}

		return result;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Bei der Anmeldung eines Empfängers für die Definitionsdaten des
	 * Skriptlaufs werden diese über den Datenverteiler publiziert. Das gleiche
	 * Verfahren trifft für die Veröffentlichung des aktuellen
	 * Ausführungszustandes des Skriptlaufs zu.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription, byte)
	 */
	@Override
	public void dataRequest(final SystemObject arg0, final DataDescription arg1, final byte arg2) {
		// Keine Aktion erfoderlich, alle Daten werden permanent zur Verfügung
		// gestellt
	}

	/**
	 * setzt die Ausführung eines angehaltenen Skripts fort.
	 *
	 * @return das Ergebnis des Kommandoausführung
	 */
	public LaufKommandoAntwort fortsetzen() {
		LaufKommandoAntwort antwort = null;
		final ServerSkript skript = getSkript();
		if (skript == null) {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
					"Dem Skriptlauf wurde kein Skript zugeordnet.");
		} else if (SkriptLaufStatus.ANGEHALTEN.equals(getStatus())) {
			if (getInterpreter() != null) {
				getInterpreter().fortsetzen();
			}
			setStatus(SkriptLaufStatus.AKTIV);
			publiziereStatus();
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.OK, "Skript wird fortgesetzt.");
		} else {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.NICHT_AKTIV, "Skript wurde nicht angehalten.");
		}

		return antwort;
	}

	/**
	 * interpretiert die Daten als Kommando zur Ausführung des Skriptlaufs.
	 *
	 * @param data
	 *            die Daten
	 */
	private void fuehreKommandoAus(final Data data) {
		assert data != null;

		final long applikation = data.getUnscaledValue("Anwendung").longValue();
		final long id = data.getUnscaledValue("Id").longValue();
		final SkriptLaufKommando kommando = SkriptLaufKommando.getZustand(data.getUnscaledValue("Kommando").intValue());
		LaufKommandoAntwort response = null;

		if (SkriptLaufKommando.STARTEN.equals(kommando)) {
			response = starten();
		} else if (SkriptLaufKommando.ANHALTEN.equals(kommando)) {
			response = anhalten();
		} else if (SkriptLaufKommando.FORTSETZEN.equals(kommando)) {
			response = fortsetzen();
		} else if (SkriptLaufKommando.BEENDEN.equals(kommando)) {
			response = beenden();
		}

		if (response == null) {
			response = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
					"Kommando konnte nicht ausgeführt werden");
		}

		final Data responseData = getModul().getVerbindung().createData(kommandoRespDesc.getAttributeGroup());
		responseData.getUnscaledValue("Anwendung").set(applikation);
		responseData.getUnscaledValue("Id").set(id);
		responseData.getUnscaledValue("Kommando").set(response.getTyp().getCode());
		responseData.getTextArray("Meldung").setLength(1);
		responseData.getArray("Meldung").getTextValue(0).setText(response.getMeldung());

		try {
			getModul().getVerbindung()
			.sendData(new ResultData(objekt, kommandoRespDesc, System.currentTimeMillis(), responseData));
		} catch (final DataNotSubscribedException e) {
			// sollte nicht auftreten, da die Anmeldung als Quelle erfolgt ist
			e.printStackTrace();
		} catch (final SendSubscriptionNotConfirmed e) {
			// sollte nicht auftreten, da die Anmeldung als Quelle erfolgt ist
			e.printStackTrace();
		}
	}

	/**
	 * erzeugt und füllt einen Datensatz mit den aktuellen Definitionsdaten des
	 * Skriptlaufs.
	 *
	 * @return den Datensatz
	 */
	private Data fuelleDatensatz() {
		final Data data = modul.getVerbindung().createData(istDesc.getAttributeGroup());
		data.getTextValue(UDA.LAUF_ATT_NAME).setText(getName());
		data.getTextValue(UDA.LAUF_ATT_BESCHREIBUNG).setText(getBeschreibung());
		final ServerSkript ref = getSkript();
		if (ref == null) {
			data.getReferenceValue(UDA.LAUF_ATT_SKRIPT).setSystemObject(null);
		} else {
			data.getReferenceValue(UDA.LAUF_ATT_SKRIPT).setSystemObject(ref.getDavObjekt());
		}

		final Data.Array parameterArray = data.getArray(UDA.LAUF_ATT_PARAMETER);
		parameterArray.setLength(getParameter().keySet().size());
		int idx = 0;
		for (final String key : getParameter().keySet()) {
			parameterArray.getItem(idx).getTextValue("Name").setText(key);
			parameterArray.getItem(idx).getTextValue("Wert").setText(getParameter().get(key));
			idx++;
		}

		data.getTextValue(UDA.LAUF_ATT_BENUTZER).setText(getBenutzer());

		final Data.NumberArray passwort = data.getUnscaledArray(UDA.LAUF_ATT_PASSWORT);
		passwort.setLength(0);

		data.getUnscaledValue(UDA.LAUF_ATT_STARTTYP).set(getStartTyp().getCode());

		data.getItem(UDA.LAUF_ATT_STARTZYKLUS).getTimeValue(UDA.ATL_STARTZYKLUS_ATT_STARTZEITPUNKT)
		.setMillis(getZyklus().getStartZeitPunkt());
		data.getItem(UDA.LAUF_ATT_STARTZYKLUS).getUnscaledValue(UDA.ATL_STARTZYKLUS_ATT_INTERVALL)
		.set(getZyklus().getIntervall());

		final Data.Array protokollArray = data.getArray(UDA.LAUF_ATT_PROTOKOLL);
		final Collection<Protokoll> protokolle = getProtokolle();
		protokollArray.setLength(protokolle.size());
		idx = 0;
		for (final Protokoll protokoll : protokolle) {
			protokollArray.getItem(idx).getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_TYP).set(protokoll.getTyp().getCode());
			protokollArray.getItem(idx).getTextValue(UDA.ATL_PROTOKOLL_ATT_NAME).setText(protokoll.getName());
			protokollArray.getItem(idx).getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_ANZAHL).set(protokoll.getAnzahl());
			protokollArray.getItem(idx).getUnscaledValue(UDA.ATL_PROTOKOLL_ATT_MAXGROESSE).set(protokoll.getMaxSize());
			idx++;
		}

		final Data.Array historie = data.getArray(UDA.LAUF_ATT_HISTORIE);
		historie.setLength(getHistorie().size());
		for (idx = 0; idx < getHistorie().size(); idx++) {
			final Historie eintrag = getHistorie().get(idx);
			historie.getItem(idx).getTimeValue(UDA.ATL_HISTORIE_ATT_AENDERUNGSDATUM).setMillis(eintrag.getZeitPunkt());
			historie.getItem(idx).getTextValue(UDA.ATL_HISTORIE_ATT_NAME).setText(eintrag.getName());
			historie.getItem(idx).getTextValue(UDA.ATL_HISTORIE_ATT_GRUND).setText(eintrag.getGrund());
		}

		return data;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.uda.data.SkriptLauf#getModul()
	 */
	@Override
	public ServerUdaModul getModul() {
		return modul;
	}

	/**
	 * liefert das Datenverteiler-Systemobjekt, das den Skriptlauf
	 * repräsentiert.
	 *
	 * @return objekt
	 */
	public SystemObject getObjekt() {
		return objekt;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion überlädt die Funktion der Basisklasse und liefert ein
	 * Skriptobjekt zurück, wie es innerhalb des Servers der "Umfassenden
	 * Datenanalyse" verwendet wird.
	 *
	 * @see ServerSkriptLauf#setSkript(Skript)
	 * @see #setSkript(Skript)
	 */
	@Override
	public ServerSkript getSkript() {
		if (super.getSkript() != null) {
			setSkript(getModul().getSkript(super.getSkript().getDavObjekt().getPid()));
		}
		return (ServerSkript) super.getSkript();
	}

	/**
	 * stellt die Verbindung zum Datenverteiler her.<br>
	 * DIe Funktion ermittelt die Datenbeschreibungen für den Versand und dem
	 * Empfang der Skriptlaufdaten und meldet das Objekt als Quelle für den
	 * Istzustand der Definitionen des Skriptlaufs und als Senke für Vorgaben
	 * zur Veränderung des Objekts an.<br>
	 * Desweiteren ist das Objekt als Quelle für den Zustand der
	 * Skriptlaufausführung angemeldet.
	 *
	 * @throws UdaServerException
	 *             die Verbindung mit dem Datenverteiler konnte nicht korrekt
	 *             hergestellt werden
	 */
	private void initialisiereVerbindung() throws UdaServerException {
		try {
			ladeDaten(laufDatei);
			final ClientDavInterface dav = modul.getVerbindung();
			final DataModel model = dav.getDataModel();

			/*
			 * Anmeldung als Quelle für den Ist-Zustand der Skriptlaufdefinition
			 */
			istDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF), model.getAspect(UDA.ASP_IST));
			dav.subscribeSender(this, objekt, istDesc, SenderRole.source());

			/*
			 * Anmeldung als Senke für Anforderungen zur Änderung der
			 * Skriptlaufdefinition
			 */
			vorgabeDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF),
					model.getAspect(UDA.ASP_VORGABE));
			dav.subscribeReceiver(this, objekt, vorgabeDesc, ReceiveOptions.normal(), ReceiverRole.drain());

			/* Anmeldung als Quelle für den Zustand der Skriptausführung */
			statusDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF_ZUSTAND),
					model.getAspect(UDA.ASP_IST));
			dav.subscribeSender(this, objekt, statusDesc, SenderRole.source());

			/* Anmeldung als Senke für Befehle zur Ausführung des Skripts */
			kommandoDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF_KOMMANDO),
					model.getAspect(UDA.ASP_ANFORDERUNG));
			dav.subscribeReceiver(this, objekt, kommandoDesc, ReceiveOptions.normal(), ReceiverRole.drain());

			/*
			 * Anmeldung als Quelle für Antworten auf Befehle zur Ausführung des
			 * Skripts
			 */
			kommandoRespDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF_KOMMANDO),
					model.getAspect(UDA.ASP_ANTWORT));
			dav.subscribeSender(this, objekt, kommandoRespDesc, SenderRole.source());

			/*
			 * Anmeldung als Quelle für Ausgaben des Skriptlaufs bei der
			 * Skriptausführung
			 */
			loggerDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF_LOGGER),
					model.getAspect(UDA.ASP_IST));
			dav.subscribeSender(this, objekt, loggerDesc, SenderRole.source());

			/*
			 * Anmeldung als Quelle für Ausgaben der Eregnisses des Skriptlaufs
			 */
			resultDesc = new DataDescription(model.getAttributeGroup(UDA.LAUF_ATG_LAUF_ERGEBNIS),
					model.getAspect(UDA.ASP_IST));
			dav.subscribeSender(this, objekt, resultDesc, SenderRole.source());

		} catch (final OneSubscriptionPerSendData e) {
			throw new UdaServerException("Skriptlauf: " + getName() + " konnte nicht angelegt werden", e);
		} catch (final IOException e) {
			throw new UdaServerException("Skriptlauf: " + getName() + " konnte nicht angelegt werden", e);
		}
	}

	/**
	 * {@inheritDoc}.<br>
	 * Das Objekt unterstützt die Sendesteuerung für alle Datenspezifikationen
	 * für die eine Anmeldung als Quelle erfolgte.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription)
	 */
	@Override
	public boolean isRequestSupported(final SystemObject arg0, final DataDescription arg1) {
		return true;
	}

	/**
	 * lädt die Daten des Objekts aus der übergebenen Datei.
	 *
	 * @param file
	 *            die Datei
	 * @throws IOException
	 *             die Daten konnten nicht eingelesen werden
	 */
	protected final void ladeDaten(final File file) throws IOException {
		if (file.exists()) {
			try (ObjectInputStream input = new ObjectInputStream(new FileInputStream(file))) {
				readExternal(input);
			} catch (final ClassNotFoundException e) {
				final String message = "Ungültiges Dateiformat: Die Daten des Skriptlaufs konnten aus der Datei: "
						+ file.getAbsolutePath() + " nicht geladen werden";
				throw new RuntimeException(message, e);
			}
		}
	}

	// /**
	// * {@inheritDoc}.<br>
	// * Die Funktion liefert den absolten Pfad der Datei in der die Daten des
	// * Skriptobjekts im Dateisystem der "Umfassenden Datenanalyse" abgelegt
	// * sind.
	// *
	// * @see de.bsvrz.ibv.uda.uda.data.SkriptLauf#getFileName()
	// */
	// @Override
	// public String getFileName() {
	// File file = File.createTempFile(null, null);
	// file.deleteOnExit();
	// file.
	// return laufDatei.getAbsolutePath();
	// }

	/**
	 * veröffentlicht die aktuellen Skriptlaufdaten über die entsprechende
	 * Datenverteiler-Attributgruppe.
	 */
	private void publiziereDaten() {
		try {
			sichereDaten();
			final Data data = fuelleDatensatz();
			try {
				modul.getVerbindung().sendData(new ResultData(objekt, istDesc, System.currentTimeMillis(), data));
			} catch (final DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
				ServerSkriptLauf.LOGGER.error(e.getLocalizedMessage(), e);
				// sollte nicht auftreten, da das Objekt als Quelle fungiert
			}
		} catch (final IOException e) {
			try {
				ServerSkriptLauf.LOGGER.error(e.getLocalizedMessage());
				ladeDaten(laufDatei);
			} catch (final IOException e1) {
				throw new RuntimeException(
						"Schwerwiegender Fehler: Schreiben und lesen von Dateien im Dateisystem nicht möglich", e1);
			}
		}
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see de.bsvrz.ibv.uda.uda.data.SkriptLauf#publiziereErgebnis(java.lang.Exception)
	 */
	@Override
	public void publiziereErgebnis(final Exception fehler) {
		if (fehler != null) {
			String msg = fehler.getMessage();
			if (msg == null) {
				msg = fehler.toString();
			}
			ausgeben(new ProtokollEintrag(msg));
		}

		final Data daten = getModul().getVerbindung().createData(resultDesc.getAttributeGroup());

		final Data.TextArray array = daten.getTextArray("Meldung");

		if (fehler == null) {
			daten.getUnscaledValue("Abgeschlossen").set(1);
			array.setLength(1);
			array.getTextValue(0).setText("Skriptlauf korrekt beendet");
		} else {
			final List<String> meldungen = new ArrayList<>();
			if (fehler instanceof UdaFehler) {
				UdaFehler udaFehler = (UdaFehler) fehler;
				while (udaFehler != null) {
					meldungen.add(
							udaFehler.getSkriptName() + "(" + udaFehler.getZeile() + "): " + udaFehler.getZusatztext());
					udaFehler = udaFehler.getVorgaenger();
				}
			} else if (fehler instanceof UdaInterpreterException) {
				meldungen.add("Parserfehler");
				final Collection<UdaParserFehler> udaFehler = ((UdaInterpreterException) fehler).getFehlerListe();
				for (final UdaParserFehler eintrag : udaFehler) {
					meldungen.add("" + eintrag.getLine() + ": " + eintrag.getMessage());
				}
			} else {
				String msg = fehler.getMessage();
				if (msg == null) {
					msg = fehler.toString();
				}
				meldungen.add(msg);
			}

			daten.getUnscaledValue("Abgeschlossen").set(0);
			array.setLength(meldungen.size());
			for (int idx = 0; idx < meldungen.size(); idx++) {
				array.getTextValue(idx).setText(meldungen.get(idx));
			}
		}

		try {
			getModul().getVerbindung().sendData(new ResultData(objekt, resultDesc, System.currentTimeMillis(), daten));
		} catch (final DataNotSubscribedException e) {
			// sollte nicht auftreten bei Anmeldung als Quelle
			e.printStackTrace();
		} catch (final SendSubscriptionNotConfirmed e) {
			// sollte nicht auftreten bei Anmeldung als Quelle
			e.printStackTrace();
		}
	}

	/**
	 * die Funktion veröffentlich den aktuellen AUsführungszustand des Skripts
	 * über die entsprechende Attributgruppe.
	 */
	private void publiziereStatus() {
		try {
			final Data data = getModul().getVerbindung().createData(statusDesc.getAttributeGroup());
			data.getUnscaledValue("Status").set(getStatus().getCode());
			modul.getVerbindung().sendData(new ResultData(objekt, statusDesc, System.currentTimeMillis(), data));
		} catch (final DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
			ServerSkriptLauf.LOGGER.error(e.getLocalizedMessage(), e);
			// sollte nicht auftreten, da das Objekt als Quelle fungiert
		}
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
	 */
	@Override
	@SuppressWarnings("unchecked")
	public void readExternal(final ObjectInput in) throws IOException, ClassNotFoundException {
		final Map<ServerSkriptLauf.SkriptLaufFeld, Object> daten = (Map<ServerSkriptLauf.SkriptLaufFeld, Object>) in
				.readObject();

		setName((String) daten.get(ServerSkriptLauf.SkriptLaufFeld.NAME));
		setBeschreibung((String) daten.get(ServerSkriptLauf.SkriptLaufFeld.BESCHREIBUNG));
		setBenutzer((String) daten.get(ServerSkriptLauf.SkriptLaufFeld.BENUTZER));
		setPasswd((String) daten.get(ServerSkriptLauf.SkriptLaufFeld.PASSWD));

		setSkript(getModul().getSkript((String) daten.get(ServerSkriptLauf.SkriptLaufFeld.SKRIPT)));

		clearParameter();
		final HashMap<String, String> parameter = (HashMap<String, String>) daten
				.get(ServerSkriptLauf.SkriptLaufFeld.SKRIPTPARAMETER);
		for (final String key : parameter.keySet()) {
			setParameter(key, parameter.get(key));
		}

		setStartTyp(StartTyp.getZustand((Integer) daten.get(ServerSkriptLauf.SkriptLaufFeld.STARTTYP)));
		setZyklus((StartZyklus) daten.get(ServerSkriptLauf.SkriptLaufFeld.ZYKLUS));

		clearProtokolle();
		final Collection<Protokoll> protokollListe = (Collection<Protokoll>) daten
				.get(ServerSkriptLauf.SkriptLaufFeld.PROTOKOLL);
		if (protokollListe != null) {
			getProtokolle().addAll(protokollListe);
		}

		clearHistorie();
		final List<Historie> historieListe = (List<Historie>) daten.get(ServerSkriptLauf.SkriptLaufFeld.HISTORIE);
		if (historieListe != null) {
			getHistorie().addAll(historieListe);
		}
	}

	/**
	 * entfernt alle extern gespeicherten Daten des Objekts.
	 *
	 * @throws ConfigurationChangeException
	 *             beim Entfernenen des Objekts aus der
	 *             Datenverteilerkonfiguration ist ein Fehler aufgetreten
	 */
	void remove() throws ConfigurationChangeException {
		laufDatei.delete();

		final ClientDavInterface dav = getModul().getVerbindung();

		dav.unsubscribeSender(this, objekt, istDesc);
		dav.unsubscribeReceiver(this, objekt, vorgabeDesc);
		dav.unsubscribeSender(this, objekt, statusDesc);
		dav.unsubscribeReceiver(this, objekt, kommandoDesc);
		dav.unsubscribeSender(this, objekt, kommandoRespDesc);
		objekt.invalidate();
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion überladt die Methode aus der Basisklasse. Es kann nur ein
	 * Skript vom Typ Servertyp gesetzt werden, da Instanzen dieser Klasse nur
	 * innerhalb der Serveranwendung der "Umfassenden Datenanalyse" verwendet
	 * werden.
	 *
	 * @see de.bsvrz.ibv.uda.uda.data.SkriptLauf#setSkript(de.bsvrz.ibv.uda.uda.data.Skript)
	 * @see #getSkript()
	 */
	@Override
	public void setSkript(final Skript skript) {
		if ((skript != null) && (!(skript instanceof ServerSkript))) {
			throw new IllegalArgumentException("Unzulässiger Skripttyp : " + skript.getClass());
		}
		super.setSkript(skript);
	}

	/**
	 * sichert die Daten des Skripts im Dateisystem der "Umfassenden
	 * Datenanalyse".
	 *
	 * @throws IOException
	 *             die Daten konnten nicht erfolgreich im Dateisystem gesichert
	 *             werden
	 */
	protected void sichereDaten() throws IOException {
		try (ObjectOutputStream output = new ObjectOutputStream(new FileOutputStream(laufDatei))) {
			writeExternal(output);
		}
	}

	/**
	 * startet den Skriptlauf.
	 *
	 * @return das Ergebnis des Kommandoausführung
	 */
	public LaufKommandoAntwort starten() {
		LaufKommandoAntwort antwort = null;
		final ServerSkript skript = getSkript();
		if (skript == null) {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
					"Dem Skriptlauf wurde kein Skript zugeordnet.");
		} else if (SkriptLaufStatus.PASSIV.equals(getStatus()) || SkriptLaufStatus.BEENDET.equals(getStatus())) {
			if (SkriptZustand.FREIGEGEBEN.equals(skript.getZustand())) {
				try {
					// Sofort auszuführende Skriptläufe werden nach dem Start
					// auf den Starttyp EXPLIZIT korrigiert, um eine mehrfache
					// Ausführung zu verhindern.
					if (StartTyp.SOFORT.equals(getStartTyp())) {
						setStartTyp(StartTyp.EXPLIZIT);
						try {
							sichereDaten();
						} catch (final IOException e) {
							// der Fehler wird an dieser Stelle ignoriert, aber
							// eine Betriebsmeldung versendet
							MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.ERROR,
									"Korrektur des Starttyps in Skriptlauf " + getName() + " nicht möglich: "
											+ e.getMessage());
						}
					}
					letzterStartZeitPunkt = System.currentTimeMillis();
					setStatus(SkriptLaufStatus.AKTIV);
					laufInterpreter = new UdaInterpreter(getModul().getVerbindung(), this);
					laufInterpreter.addListener(this);
					laufInterpreter.start();
					publiziereStatus();
					antwort = new LaufKommandoAntwort(SkriptLaufKommando.OK, "Skript wird gestartet.");
				} catch (final UdaInterpreterException e) {
					antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER, e.toString());
					setStatus(SkriptLaufStatus.BEENDET);
					publiziereDaten();
				}
			} else {
				antwort = new LaufKommandoAntwort(SkriptLaufKommando.ALLGEMEINER_FEHLER,
						"Skript wurde noch nicht freigegeben.");
			}
		} else {
			antwort = new LaufKommandoAntwort(SkriptLaufKommando.SCHON_AKTIV, "Skriptlauf ist bereits aktiv.");
		}

		return antwort;
	}

	/**
	 * {@inheritDoc}.<br>
	 * Die Funktion sorgt für die Verarbeitung der empfangenen Vorgaben für die
	 * Änderungen innerhalb einer Skriptlaufdefinition.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public void update(final ResultData[] resultListe) {
		for (final ResultData result : resultListe) {
			final Data data = result.getData();
			if (data != null) {
				final AttributeGroup atg = result.getDataDescription().getAttributeGroup();
				if (atg.equals(vorgabeDesc.getAttributeGroup())) {
					try {
						aktualisiereDaten(data);
						sichereDaten();
						publiziereDaten();
					} catch (final IOException e) {
						try {
							ServerSkriptLauf.LOGGER.error(e.getLocalizedMessage());
							ladeDaten(laufDatei);
						} catch (final IOException e1) {
							LogTools.log(modul.getLogger(), UdaMessage.ErrorDateisystem, laufDatei.getAbsolutePath(),
									e1.getMessage());
							throw new RuntimeException(MessageFormat.format(UdaMessage.ErrorDateisystem.toString(),
									laufDatei.getAbsolutePath(), e1.getMessage()));
						}
						publiziereDaten();
					} catch (final UdaServerException e) {
						// Fehler wird an dieser Stelle ignoriert
						LogTools.log(modul.getLogger(), UdaMessage.ErrorSkriptlaufAktualisierung, getName(),
								e.getMessage());
					}
				} else if (atg.equals(kommandoDesc.getAttributeGroup())) {
					fuehreKommandoAus(data);
				}
			}
		}
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
	 */
	@Override
	public void writeExternal(final ObjectOutput out) throws IOException {

		final Map<ServerSkriptLauf.SkriptLaufFeld, Object> daten = new HashMap<>();

		daten.put(ServerSkriptLauf.SkriptLaufFeld.NAME, getName());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.BESCHREIBUNG, getBeschreibung());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.BENUTZER, getBenutzer());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.PASSWD, getPasswd());

		if (getSkript() != null) {
			daten.put(ServerSkriptLauf.SkriptLaufFeld.SKRIPT, getSkript().getDavObjekt().getPid());
		}

		daten.put(ServerSkriptLauf.SkriptLaufFeld.SKRIPTPARAMETER, getSkriptParameter());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.STARTTYP, getStartTyp().getCode());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.ZYKLUS, getZyklus());
		daten.put(ServerSkriptLauf.SkriptLaufFeld.HISTORIE, getHistorie());

		daten.put(ServerSkriptLauf.SkriptLaufFeld.PROTOKOLL, getProtokolle());

		out.writeObject(daten);
	}
}
