/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.3 Umfassende Datenanalyse
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.uda.interpreter.daten.fuzzy;

import java.util.*;

import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.ArgumentFehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.Fehler;
import de.bsvrz.ibv.uda.interpreter.daten.fehler.UdaFehlerSubtyp;
import de.bsvrz.sys.funclib.bitctrl.interpreter.*;
import de.bsvrz.sys.funclib.debug.Debug;

import static de.bsvrz.ibv.uda.interpreter.daten.fuzzy.UdaFuzzyOperators.*;

/**
 * Handler für die Ausführung von Fuzzy-Opertaionen innerhalb eines
 * UdaInterpreter.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class UdaFuzzyHandler extends Handler {

	private static final Debug LOGGER = Debug.getLogger();

	@Override
	public Operator[] getHandledOperators() {
		return UdaFuzzyOperators.OPERATOREN;
	}

	@Override
	public Object perform(final Operator operator, final List<?> operanden) {
		final HandlerValidation validation = validiereHandler(operator, operanden);
		if (!validation.isValid()) {
			if (!validation.isRichtigeAnzahl()) {
				throw new ArgumentFehler(UdaFehlerSubtyp.ANZAHL);
			}
			if (!validation.isRichtigerTyp()) {
				throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE);
			}
			throw new ArgumentFehler(UdaFehlerSubtyp.UNBEKANNT);
		}

		Object result = null;
		try {
			if (LING_TERM_KONSTRUKTOR.equals(operator)) {
				final String name = (String) operanden.get(0);
				final Number unten = (Number) operanden.get(1);
				final Number oben = (Number) operanden.get(2);
				String desc = null;
				if (operanden.size() > 3) {
					desc = (String) operanden.get(3);
				}
				result = new UdaLinguistischerTerm(name, unten.intValue(), oben.intValue(), desc);
			} else if (LING_VAR_KONSTRUKTOR.equals(operator)) {
				final String name = (String) operanden.get(0);
				final Number unten = (Number) operanden.get(1);
				final Number oben = (Number) operanden.get(2);
				final Collection<UdaLinguistischerTerm> terme = new ArrayList<>();
				for (int idx = 3; idx < operanden.size(); idx++) {
					terme.add((UdaLinguistischerTerm) operanden.get(idx));
				}
				result = new UdaLinguistischeVariable(name, unten.intValue(), oben.intValue());
				for (final UdaLinguistischerTerm set : terme) {
					((UdaLinguistischeVariable) result).addLinguistischerTerm(set);
				}
			} else if (FUZZYFIZIERE.equals(operator)) {
				final UdaLinguistischeVariable var = (UdaLinguistischeVariable) operanden.get(0);
				if (operanden.size() > 1) {
					final Number wert = (Number) operanden.get(1);
					result = new UdaFuzzyWert(var.getLinguistischeVariable().fuzzyfiziere(wert.intValue()));
				} else {
					result = new UdaFuzzyWert(var.getLinguistischeVariable().erzeugeFuzzyVariable());
				}
			} else if (DEFUZZYFIZIERE.equals(operator)) {
				final UdaLinguistischeVariable var = (UdaLinguistischeVariable) operanden.get(0);
				final UdaFuzzyWert wert = (UdaFuzzyWert) operanden.get(1);
				result = var.getLinguistischeVariable().defuzzyfiziere(wert.getVariable());
				if (Double.isNaN((Double) result)) {
					throw new Fehler("Die Defuzzyfizierung konnte nicht ausgeführt werden");
				}
			} else if (FUZZYWERT.equals(operator)) {
				final UdaLinguistischeVariable var = (UdaLinguistischeVariable) operanden.get(0);
				result = new UdaFuzzyWert(var.getLinguistischeVariable().erzeugeFuzzyVariable());
			} else if (ZUGEHOERIGKEIT.equals(operator)) {
				final UdaFuzzyWert wert = (UdaFuzzyWert) operanden.get(0);
				final String name = (String) operanden.get(1);
				if (operanden.size() <= 2) {
					final Term term = (Term) wert.getStrukturElement(name);
					if (term == null) {
						throw new ArgumentFehler(UdaFehlerSubtyp.ARGUMENTE,
								"Es existiert kein Term mit dem Namen \"" + name + "\"");
					}
					result = term.getZugehoerigkeit().getWert();
				} else {
					final Number zahl = (Number) operanden.get(2);
					wert.setStrukturElement(name, zahl);
					result = wert;
				}
			} else if (KOMPLEMENT.equals(operator)) {
				result = komplement((Number) operanden.get(0));
			} else if (GAMMA.equals(operator)) {
				result = gammaOperator((Number) operanden.get(0), (Number) operanden.get(1), (Number) operanden.get(2));
			} else if (MAXIMUM.equals(operator)) {
				result = maximum(operanden.toArray(new Number[0]));
			} else if (MINIMUM.equals(operator)) {
				result = minimum(operanden.toArray(new Number[0]));
			} else if (ALGSUMME.equals(operator)) {
				result = algebraischeSumme((Number) operanden.get(0), (Number) operanden.get(1));
			} else if (ALGPRODUKT.equals(operator)) {
				result = algebraischesProdukt((Number) operanden.get(0), (Number) operanden.get(1));
			}
		} catch (final ClassCastException e) {
			final String message = "Der Operator \"" + operator.getSymbol() + "\" ist auf den Typ \""
					+ operanden.get(0).getClass() + "\" nicht anwendbar";
			UdaFuzzyHandler.LOGGER.error(message, e);
			throw new Fehler(message);
		}
		return result;
	}

	@Override
	public HandlerValidation validiereHandler(Operator operator, List<?> operanden) {
		boolean anzahlOk = validiereAnzahl(operator, operanden);
		boolean typOk = validiereTyp(operator, operanden);
		return new HandlerValidation(anzahlOk, typOk);
	}

	private static boolean validiereAnzahl(Operator operator, List<?> operanden) {
		if (MINIMUM.equals(operator) || MAXIMUM.equals(operator)) {
			return operanden.size() >= 2;
		} else if (KOMPLEMENT.equals(operator)) {
			return operanden.size() == 1;
		} else if (ALGPRODUKT.equals(operator) || ALGSUMME.equals(operator)) {
			return operanden.size() == 2;
		} else if (GAMMA.equals(operator)) {
			return operanden.size() == 3;
		} else if (LING_TERM_KONSTRUKTOR.equals(operator) ||
				LING_VAR_KONSTRUKTOR.equals(operator) ||
				FUZZYFIZIERE.equals(operator) ||
				DEFUZZYFIZIERE.equals(operator) ||
				FUZZYWERT.equals(operator) ||
				ZUGEHOERIGKEIT.equals(operator)
		) {
			return true;
		} else {
			return false;
		}
	}

	private static boolean validiereTyp(Operator operator, List<?> operanden) {
		if (LING_TERM_KONSTRUKTOR.equals(operator) ||
				LING_VAR_KONSTRUKTOR.equals(operator) ||
				FUZZYFIZIERE.equals(operator) ||
				DEFUZZYFIZIERE.equals(operator) ||
				FUZZYWERT.equals(operator) ||
				ZUGEHOERIGKEIT.equals(operator)
		) {
			return true;
		} else {
			return operanden.stream()
					.map(o -> o instanceof Number)
					.reduce(Boolean::logicalAnd)
					.orElse(false);
		}
	}

	private static Object minimum(Number... operanden) {
		return minimum(Arrays.asList(operanden));
	}

	private static Object minimum(Collection<Number> operanden) {
		Optional<Double> z = operanden.stream()
				.map(Number::doubleValue)
				.map(Zugehoerigkeit::vonNullwert)
				.reduce(Funktionen::minimum)
				.map(Zugehoerigkeit::getWert);
		if (z.isPresent()) {
			return z.get();
		} else {
			return NichtWert.NICHTWERT;
		}
	}


	private static Object maximum(Number... operanden) {
		return maximum(Arrays.asList(operanden));
	}

	private static Object maximum(Collection<Number> operanden) {
		Optional<Double> z = operanden.stream()
				.map(Number::doubleValue)
				.map(Zugehoerigkeit::vonNullwert)
				.reduce(Funktionen::maximum)
				.map(Zugehoerigkeit::getWert);
		if (z.isPresent()) {
			return z.get();
		} else {
			return NichtWert.NICHTWERT;
		}
	}

	private static Object komplement(Number a) {
		Zugehoerigkeit za = Zugehoerigkeit.vonNullwert(a.doubleValue());
		Zugehoerigkeit z = Funktionen.komplement(za);
		if (z.nonNull()) {
			return z.getWert();
		} else {
			return NichtWert.NICHTWERT;
		}
	}

	private static Object algebraischesProdukt(Number a, Number b) {
		Zugehoerigkeit za = Zugehoerigkeit.vonNullwert(a.doubleValue());
		Zugehoerigkeit zb = Zugehoerigkeit.vonNullwert(b.doubleValue());
		Zugehoerigkeit z = Funktionen.algebraischesProdukt(za, zb);
		if (z.nonNull()) {
			return z.getWert();
		} else {
			return NichtWert.NICHTWERT;
		}
	}

	private static Object algebraischeSumme(Number a, Number b) {
		Zugehoerigkeit za = Zugehoerigkeit.vonNullwert(a.doubleValue());
		Zugehoerigkeit zb = Zugehoerigkeit.vonNullwert(b.doubleValue());
		Zugehoerigkeit z = Funktionen.algebraischesSumme(za, zb);
		if (z.nonNull()) {
			return z.getWert();
		} else {
			return NichtWert.NICHTWERT;
		}
	}

	private static Object gammaOperator(Number gamma, Number a, Number b) {
		Zugehoerigkeit za = Zugehoerigkeit.vonNullwert(a.doubleValue());
		Zugehoerigkeit zb = Zugehoerigkeit.vonNullwert(b.doubleValue());
		Zugehoerigkeit zgamma = Zugehoerigkeit.vonNullwert(gamma.doubleValue());
		Zugehoerigkeit z = Funktionen.gammaOperator(za, zb, zgamma);
		if (z.nonNull()) {
			return z.getWert();
		} else {
			return NichtWert.NICHTWERT;
		}
	}

}
